;
; Inno Setup Installation script for Uni-Verse.
;
; Note: This script happily assumes that you have the various tools and data files
; under a toplevel C:\projects directory. If that is not true on your system, do not
; expect to be able to run this script without (heavy) modifications.
;

[Setup]
AppName=Uni-Verse
AppVerName=Uni-Verse Release Two (Core r6p1)
AppPublisher=Uni-Verse Consortium
AppPublisherURL=http://www.uni-verse.org/
DefaultDirName={pf}\Uni-Verse-r6p1
DefaultGroupName=Uni-Verse
;UninstallDisplayIcon={app}\MyProg.exe
Compression=lzma
;SolidCompression=yes
SetupIconFile=resources\universe-icon.ico
WizardImageFile=resources\universe-wizard.bmp

[Code]
var pageMaxDir: TInputDirWizardPage;

// Returns the last position that holds the character 'ch'. Returns -1 if not found.
function LastPosChar(str: String; ch : Char): Integer;
var
  i : Longint;
begin
  result := -1;
  for i := length(str) downto 1 do begin
    if Copy(str, i, 1) = ch then begin
      result := i;
      exit;
    end;
  end;
end;

// Attempt to locate directory where 3ds Max is installed, by going through
// the registry information for the default icon.
// Based on <http://www.scriptspot.com/Tutorials/NSIS/Creation%20Application%20Installers%20for%20MaxScripts.htm>.
function FindMaxDir() : String;
var
  x : String;
  p : Integer;
begin
  Result := '';
  if RegQueryStringValue(HKEY_CLASSES_ROOT, '3dsmax8', 'DefaultIcon', x) then begin
    p := LastPosChar(x, '\')
    if p > 4 then
      result := Copy(x, 1, p - 1);
  end;
end;

procedure InitializeWizard();
var
  md : String;
begin
  pageMaxDir := CreateInputDirPage(wpSelectComponents, 'Select 3ds Max location',
                              'Tell the installer where your copy of Autodesk''s 3ds Max is installed, so that the Verse plug-in can be properly added to the application.',
                              'Where is the 3ds Max root directory located? This installer will look for version 8 of 3ds Max, but should also work with version 7. ' +
                              'If you have version 7 installed, please manually supply the path to the root directory below.', False, '');
  pageMaxDir.add('3ds Max Root Directory');
  md := FindMaxDir();
  if length(md) > 0 then
    pageMaxDir.Values[0] := md
  else
    pageMaxDir.Values[0] := AddBackslash(ExpandConstant('{pf}')) + '3dsmax8';
end;

function ShouldSkipPage(PageID: Integer): Boolean;
begin
  Result := False;
  if PageID = pageMaxDir.ID then Result := not IsComponentSelected('maxplug');
end;

function NextButtonClick(CurPageID: Integer): Boolean;
begin
  Result := True; // Default is to permit progress, safer in case of error below.
  if IsComponentSelected('maxplug') and (CurPageID = pageMaxDir.ID) then begin
    if DirExists(pageMaxDir.Values[0]) and DirExists(AddBackslash(pageMaxDir.Values[0]) + 'plugins') then begin
      result := true
    end
    else
    begin
      MsgBox('The 3ds Max root you supply must contain a plugins\ sub-directory. '+
             'The installer cannot find one in the currently entered directory. '+
             'Please specify a different 3ds Max root.', mbInformation, MB_OK);
      result := false;
    end;
  end;
end;

// This gets called from a {code} constant, below. It returns the string given on the custom page.
function GetMaxDir(param : String) : String;
begin
  result := pageMaxDir.Values[0];
end;

// ---------------------------------------------------------------------------------------------------------------------

[Types]
Name: "full"; Description: "Full installation, without 3ds max plug-in"
Name: "full3ds"; Description: "Full installation, with 3ds max plug-in"
Name: "compact"; Description: "Compact installation"
Name: "custom"; Description: "Custom installation"; Flags: iscustom

[Components]
Name: server; Description: "Verse server"; Types: full full3ds compact custom;
Name: doc_server; Description: "Documentation for the Verse server program"; Types: full full3ds;
Name: dev; Description: "Files for developing new Verse clients"; Types: full full3ds;
Name: connector; Description: "Connector, an application that gives you schematic access to data stored in a Verse server"; Types: full full3ds compact custom;
Name: doc_connector; Description: "Documentation for Connector"; Types: full full3ds custom;
Name: loqairou; Description: "Loq Airou, a subdivision surfaces-based online modelling tool"; Types: full full3ds compact custom;
Name: doc_loqairou; Description: "Documentation for Loq Airou"; Types: full full3ds custom;
Name: quelsolaar; Description: "Quel Solaar, a high-qualty Verse rendering client"; Types: full full3ds compact custom;
Name: uvedit; Description: "UVEdit, a tool for editing UV-mappings, used when mapping a flat texture onto the surface of a 3D object"; Types: full full3ds custom;
Name: lpaint; Description: "Layer Painter, a tool for painting values onto Verse geometry layers"; Types: full full3ds;
Name: render; Description: "Simple Renderer; a simplistic rendering client useful for quick low-quality visualization"; Types: full full3ds;
Name: doc_render; Description: "Documentation for the Simple Renderer"; Types: full full3ds;
Name: osgverse; Description: "OSGVerse, a Verse rendering application based on the OpenSG toolkit"; Types: full;
Name: loader; Description: "Loader, a command-line tool for loading Verse data from files in the VML format"; Types: full full3ds compact custom;
Name: saver; Description: "Saver, a command-line tool for saving the state of a Verse server as a file in the VML format"; Types: full full3ds compact custom;
Name: doc_vml; Description: "Documentation for the VML loader/saver tools"; Types: full full3ds;
Name: vml; Description: "Sample VML files, for use with the loader tool"; Types: full full3ds;
Name: purple; Description: "Purple, a data-flow based computation engine for Verse"; Types: full full3ds custom;
Name: doc_purple; Description: "Internet web link for Purple documentation"; Types: full full3ds custom;
Name: puyo; Description: "Puyo, a user-interface client for Purple"; Types: full full3ds custom;
Name: doc_puyo; Description: "Internet web link for Puyo user interface client documentation"; Types: full full3ds custom;
Name: maxplug; Description: "Verse plug-in for Autodesk 3ds Max"; Types: full3ds;

[Files]
Source: "c:\projects\verse\verse.exe"; DestDir: "{app}"; Components: server;
Source: "c:\projects\verse\verse.h"; DestDir: "{app}"; Components: dev;
Source: "c:\projects\verse\verse.lib"; DestDir: "{app}"; Components: dev;
Source: "c:\projects\verse\README.html"; DestDir: "{app}\docs\server"; Components: doc_server;
Source: "c:\projects\quelsolaar\connector.exe"; DestDir: "{app}"; Components: connector;
Source: "c:\projects\quelsolaar\docs\connector\*"; DestDir: "{app}\docs\connector"; Components: doc_connector;
Source: "c:\projects\quelsolaar\loqairou.exe"; DestDir: "{app}"; Components: loqairou;
Source: "c:\projects\quelsolaar\docs\loq-airou\*"; DestDir: "{app}\docs\loqairou"; Components: doc_loqairou;
Source: "c:\projects\quelsolaar\quelsolaar.exe"; DestDir: "{app}"; Components: quelsolaar;
Source: "c:\projects\quelsolaar\uvedit.exe"; DestDir: "{app}"; Components: uvedit;
Source: "c:\projects\quelsolaar\lpaint.exe"; DestDir: "{app}"; Components: lpaint;

[Files]
Source: "c:\projects\verse-tests\render.exe"; DestDir: "{app}"; Components: render;
Source: "c:\windows\system32\iconv.dll"; DestDir: "{app}"; Components: render;
Source: "c:\windows\system32\intl.dll"; DestDir: "{app}"; Components: render;
Source: "c:\windows\system32\LIBGLIB-2.0-0.DLL"; DestDir: "{app}"; Components: render;
Source: "c:\projects\verse-tests\docs\render\*.html"; DestDir: "{app}\docs\render"; Components: doc_render;
Source: "c:\projects\verse-tests\docs\render\media\*.png"; DestDir: "{app}\docs\render\media"; Components: doc_render;
Source: "c:\windows\system32\glut32.dll"; DestDir: "{app}"; Components: render;

[Files]
Source: "c:\projects\OSGVerse\*.*"; DestDir: "{app}\OSGVerse"; Components: osgverse;

[Files]
Source: "c:\projects\vml\loader.exe"; DestDir: "{app}"; Components: loader;
Source: "c:\projects\vml\saver.exe"; DestDir: "{app}"; Components: saver;
Source: "c:\projects\vml\README.html"; DestDir: "{app}\docs\vml"; Components: doc_vml;
Source: "c:\projects\vml\tests\*.vml"; DestDir: "{app}\vml"; Components: vml;
Source: "c:\projects\vml\scenes\*.vml"; DestDir: "{app}\vml"; Components: vml;
Source: "c:\projects\purple\purple.exe"; DestDir: "{app}\purple"; Components: purple;
Source: "c:\projects\purple\purple.h"; DestDir: "{app}\purple"; Components: purple;
Source: "c:\projects\purple\purple.lib"; DestDir: "{app}\purple"; Components: purple;
Source: "c:\projects\purple\plugins\*.dll"; DestDir: "{app}\purple\plugins"; Components: purple;
Source: "c:\projects\puyo\*.py"; DestDir: "{app}\puyo"; Components: puyo;

[Files]
;Source: "c:\projects\verse4max\Verse040.gup"; DestDir: "{code:GetMaxDir}\plugins"; Components: maxplug;

[Icons]
Name: "{group}\Verse Server"; Filename: "{app}\verse.exe"; WorkingDir: "{app}"; Components: server; Comment: "Start a new Verse server";
Name: "{group}\Connector"; Filename: "{app}\connector.exe"; WorkingDir: "{app}"; Components: connector; Comment: "Start Connector, the schematic Verse data explorer.";
Name: "{group}\Loq Airou"; Filename: "{app}\loqairou.exe"; WorkingDir: "{app}"; Components: loqairou; Comment: "Start Loq Airou, the 3D modelling tool.";
Name: "{group}\Quel Solaar"; Filename: "{app}\quelsolaar.exe"; WorkingDir: "{app}"; Components: quelsolaar; Comment: "Start Quel Solaar, the Verse rendering client.";
Name: "{group}\Layer Painter"; Filename: "{app}\lpaint.exe"; WorkingDir: "{app}"; Components: lpaint; Comment: "Start Layer Painter, the geometry layer painting tool.";
Name: "{group}\UV Edit"; Filename: "{app}\uvedit.exe"; WorkingDir: "{app}"; Components: uvedit; Comment: "Start UV Edit, the mapping tool.";
Name: "{group}\Simple Renderer"; Filename: "{app}\render.exe"; WorkingDir: "{app}"; Components: render; Comment: "Start the Simple Renderer rendering client.";
Name: "{group}\OSGVerse"; Filename: "{app}\OSGVerse\OSGVerse.exe"; WorkingDir: "{app}\OSGVerse"; Components: osgverse; Comment: "Start the OSGVerse rendering client.";
Name: "{group}\Purple"; Filename: "{app}\purple\purple.exe"; WorkingDir: "{app}"; Components: purple; Comment: "Start Purple, the data-flow based computation engine for Verse.";
Name: "{group}\Puyo"; Filename: "{app}\puyo\puyo.py"; WorkingDir: "{app}"; Components: purple; Comment: "Start Puyo, the user interface client for Purple."
Name: "{group}\Content\Upload material test scene to localhost"; Filename: "{app}\loader.exe"; Parameters: "material-test.vml"; WorkingDir: "{app}\vml\"; Components: loader and vml; Comment: "Upload a sample scene to server running on local machine.";
Name: "{group}\Content\Upload car test scene to localhost"; Filename: "{app}\loader.exe"; Parameters: "car.vml"; WorkingDir: "{app}\vml\"; Components: loader and vml; Comment: "Upload a sample scene to server running on local machine.";
Name: "{group}\Content\Upload Jorma bird scene to localhost"; Filename: "{app}\loader.exe"; Parameters: "jorma.vml";WorkingDir: "{app}\vml\"; Components: loader and vml; Comment: "Upload a sample scene to server running on local machine";
Name: "{group}\Content\Upload cube on ground to localhost"; Filename: "{app}\loader.exe"; Parameters: "cube-on-ground.vml";WorkingDir: "{app}\vml\"; Components: loader and vml; Comment: "Upload a sample scene to server running on local machine";

[Icons]
Name: "{group}\Documentation\Uni-Verse Documentation Wiki (web)"; Filename: "http://wiki.blender.org/index.php/Uni-Verse:Main"; Comment: "Open web page that holds Uni-Verse user documentation.";
Name: "{group}\Documentation\Verse Server"; Filename: "{app}\docs\server\README.html"; Components: server; Comment: "Open the (brief) documentation for the Verse server.";
Name: "{group}\Documentation\Connector"; Filename: "{app}\docs\connector\index.html"; Components: doc_connector; Comment: "Open documentation for Connector."; IconFilename: "c:\projects\universe-install\connector.ico";
Name: "{group}\Documentation\Loq Airou"; Filename: "{app}\docs\loqairou\index.html"; Components: doc_loqairou; Comment: "Open documentation for Loq Airou.";
Name: "{group}\Documentation\Loader&Saver"; Filename: "{app}\docs\vml\README.html"; Components: doc_vml; Comment: "Open documentation for the VML Loader and Saver tools.";
Name: "{group}\Documentation\Simple Renderer"; Filename: "{app}\docs\render\index.html"; Components: doc_render; Comment: "Open documentation for the Simple Renderer.";
Name: "{group}\Documentation\Purple"; Filename: "http://purple.blender.org/"; Components: doc_purple; Comment: "Open the Purple web site (on the Internet).";
Name: "{group}\Documentation\Puyo"; Filename: "http://purple.blender.org/cms/fileadmin/purple/puyo-guide/index.html"; Components: "doc_puyo"; Comment: "Open documentation for the Purple user interface client.";
Name: "{group}\Documentation\3ds Max Plug-in"; Filename: "http://www.pdc.kth.se/~slas/Verse4Max/doc/Verse4Max.html"; Components: maxplug; Comment: "Open Verse4Max 3ds Max plugin documentation (online link).";

[Icons]
Name: "{group}\Feedback\Send Feedback E-Mail"; Filename: "http://projects.blender.org/mailman/listinfo/universe-feedback"; Comment: "Visits the universe-feedback mailing list info page. Please subscribe, and send questions.";
; This removes some of the temporary/transient files created during the install. This
; helps make sure the target directory, {app}, is really empty, and thus that it is
; removed too. FIXME: We might want the host_id.rsa to hang around ... Not sure.
[UninstallDelete]
Type: files; Name: "{app}\*.tga";
Type: files; Name: "{app}\host_id.rsa";
Type: files; Name: "{app}\address.bin";
Type: filesandordirs; Name: "{app}\puyo";

