/*
** Initial test of the geometry node. Handy during development of said node. This program
** doesn't really do anything, it just creates a geometry node and populates it with some
** sample data, while subscribing to the same node and printing out the responses it gets
** from the server. It might be instructive as a reasonably simple, clean, Verse program.
**
** "Ported" forward from Verse generation 1. Stripped out glib, changed layer semantics.
*/

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <unistd.h>

/* Bring in the Verse header, to use the library's functions. */
#include "verse.h"

/* Some windows plumbing. */
#if !defined(M_PI)
#define M_PI	3.14159265358979323846
#endif
#if !defined(M_PI2)
#define	M_PI2	1.57079632679489661923
#endif

/* ----------------------------------------------------------------------------------------------------------------------- */

static boolean	do_build = FALSE;

static VNodeID	the_avatar = 0;

/* ----------------------------------------------------------------------------------------------------------------------- */

static void cb_g_vertex_set_xyz(void *user, VNodeID node_id, VLayerID layer_id, uint32 vertex_id, real64 x, real64 y, real64 z)
{
	printf("Node %u: Vertex %u.%u set to (%f,%f,%f)\n", node_id, layer_id, vertex_id, x, y, z);

	if(layer_id == 0 && vertex_id == 7)
		verse_send_g_layer_create(node_id, 0, "another xyz", VN_G_LAYER_VERTEX_XYZ, 0, 0.0);
}

static void cb_g_vertex_set_uint32(void *user, VNodeID node_id, VLayerID layer_id, uint32 vertex_id, uint32 value)
{
	printf("Node %u: Vertex %u.%u integer set to %u\n", node_id, layer_id, vertex_id, value);
}

static void cb_g_vertex_set_real64(void *user, VNodeID node_id, VLayerID layer_id, uint32 vertex_id, real64 value)
{
	static unsigned int	counter = 0U;

	printf("Node %u: Vertex %u.%u real set to %g\n", node_id, layer_id, vertex_id, value);

	if(do_build && ++counter == 8)
	{
		printf("All vertices uploaded, creating polygon fragment\n");

		verse_send_g_layer_create(node_id, -1, "p_base",       VN_G_LAYER_POLYGON_CORNER_UINT32, 0, 0.0);
		verse_send_g_layer_create(node_id, -2, "p_crease",     VN_G_LAYER_POLYGON_CORNER_UINT32, 0, 0.0);
		verse_send_g_layer_create(node_id, -3, "p_cornerreal", VN_G_LAYER_POLYGON_CORNER_REAL,   0, 0.0);
		verse_send_g_layer_create(node_id, -4, "p_faceint8",   VN_G_LAYER_POLYGON_FACE_UINT8,    0, 0.0);
		verse_send_g_layer_create(node_id, -5, "p_faceint32",  VN_G_LAYER_POLYGON_FACE_UINT32,   0, 0.0);
		verse_send_g_layer_create(node_id, -6, "p_facereal",   VN_G_LAYER_POLYGON_FACE_REAL,     0, 0.0);
	}
}

static void cb_g_polygon_set_corner_uint32(void *user, VNodeID node_id, VLayerID layer_id, uint32 polygon_id, uint32 v0, uint32 v1, uint32 v2, uint32 v3)
{
 	printf("Node %u: Polygon %u.%u corner integer set to %u,%u,%u,%u\n", node_id, layer_id, polygon_id, v0, v1, v2, v3);
}

static void cb_g_polygon_set_corner_real64(void *user, VNodeID node_id, VLayerID layer_id, uint32 polygon_id, real64 v0, real64 v1, real64 v2, real64 v3)
{
	printf("Node %u: Polygon %u.%u corner real set to %g,%g,%g,%g\n", node_id, layer_id, polygon_id, v0, v1, v2, v3);
}

static void cb_g_polygon_set_face_uint8(void *user, VNodeID node_id, VLayerID layer_id, uint32 polygon_id, uint8 value)
{
	printf("Node %u: Polygon %u.%u face integer8 set to %u\n", node_id, layer_id, polygon_id, value);
}

static void cb_g_polygon_set_face_uint32(void *user, VNodeID node_id, VLayerID layer_id, uint32 polygon_id, uint32 value)
{
	printf("Node %u: Polygon %u.%u face integer32 set to %u\n", node_id, layer_id, polygon_id, value);
}

static void cb_g_polygon_set_face_real64(void *user, VNodeID node_id, VLayerID layer_id, uint32 polygon_id, real64 value)
{
	printf("Node %u: Polygon %u.%u face real set to %g\n", node_id, layer_id, polygon_id, value);
}

static void cb_g_layer_create(void *user, VNodeID node_id, VLayerID id, const char *name, VNGLayerType type,
			      uint32 def_integer, real64 def_real)
{
	printf("A layer called '%s' has type %u, def_i=%u def_r=%g, id=%u [subscribing]\n", name, type, def_integer, def_real, id);

	verse_send_g_layer_subscribe(node_id, id, VN_FORMAT_REAL64);

	if(type == VN_G_LAYER_VERTEX_XYZ)
	{
		if(do_build)
		{
			verse_send_g_vertex_set_xyz_real32(node_id, id, 0, -1.0f,  1.0f, -1.0f);
			verse_send_g_vertex_set_xyz_real32(node_id, id, 1,  1.0f,  1.0f, -1.0f);
			verse_send_g_vertex_set_xyz_real32(node_id, id, 2,  1.0f,  1.0f,  1.0f);
			verse_send_g_vertex_set_xyz_real32(node_id, id, 3, -1.0f,  1.0f,  1.0f);

			verse_send_g_vertex_set_xyz_real32(node_id, id, 4, -1.0f, -1.0f, -1.0f);
			verse_send_g_vertex_set_xyz_real32(node_id, id, 5,  1.0f, -1.0f, -1.0f);
			verse_send_g_vertex_set_xyz_real32(node_id, id, 6,  1.0f, -1.0f,  1.0f);
			verse_send_g_vertex_set_xyz_real32(node_id, id, 7, -1.0f, -1.0f,  1.0f);
			printf("Vertices sent\n");
		}
		else
			printf("Not setting any XYZ values\n");
	}
	else if(type == VN_G_LAYER_VERTEX_UINT32)
	{
		if(do_build)
		{
			verse_send_g_vertex_set_uint32(node_id, id, 0, 0x4711);
			verse_send_g_vertex_set_uint32(node_id, id, 1, 0x4712);
			verse_send_g_vertex_set_uint32(node_id, id, 2, 0x4713);
			verse_send_g_vertex_set_uint32(node_id, id, 3, 0x4714);
			verse_send_g_vertex_set_uint32(node_id, id, 4, 0x4715);
			verse_send_g_vertex_set_uint32(node_id, id, 5, 0x4716);
			verse_send_g_vertex_set_uint32(node_id, id, 6, 0x4717);
			verse_send_g_vertex_set_uint32(node_id, id, 7, 0x4718);
		}
	}
	else if(type == VN_G_LAYER_VERTEX_REAL)
	{
		if(do_build)
		{
			verse_send_g_vertex_set_real64(node_id, id, 0,  M_PI);
			verse_send_g_vertex_set_real64(node_id, id, 1, -M_PI);
			verse_send_g_vertex_set_real64(node_id, id, 2,  M_PI2);
			verse_send_g_vertex_set_real64(node_id, id, 3, -M_PI2);
			verse_send_g_vertex_set_real64(node_id, id, 4,  M_PI);
			verse_send_g_vertex_set_real64(node_id, id, 5, -M_PI);
			verse_send_g_vertex_set_real64(node_id, id, 6,  M_PI2);
			verse_send_g_vertex_set_real64(node_id, id, 7, -M_PI2);
		}
	}
	else if(type == VN_G_LAYER_POLYGON_CORNER_UINT32)
	{
		if(do_build)
		{
			if(id == 0U)
			{
				verse_send_g_polygon_set_corner_uint32(node_id, id, 0,  0, 1, 2, 3);
				verse_send_g_polygon_set_corner_uint32(node_id, id, 1,  2, 1, 5, 6);
				verse_send_g_polygon_set_corner_uint32(node_id, id, 2,  7, 6, 5, 4);
				verse_send_g_polygon_set_corner_uint32(node_id, id, 3,  0, 3, 7, 4);
				verse_send_g_polygon_set_corner_uint32(node_id, id, 4,  3, 2, 6, 7);
				verse_send_g_polygon_set_corner_uint32(node_id, id, 5,  1, 0, 4, 5);
			}
			else
			{
				verse_send_g_polygon_set_corner_uint32(node_id, id, 0,   0,  1,  2,  3);
				verse_send_g_polygon_set_corner_uint32(node_id, id, 1,   4,  5,  6,  7);
				verse_send_g_polygon_set_corner_uint32(node_id, id, 2,   8,  9, 10, 11);
				verse_send_g_polygon_set_corner_uint32(node_id, id, 3,  12, 13, 14, 15);
				verse_send_g_polygon_set_corner_uint32(node_id, id, 4,  16, 17, 18, 19);
				verse_send_g_polygon_set_corner_uint32(node_id, id, 5,  20, 21, 22, 23);
			}
			printf("Polygons sent\n");
		}
	}
	else if(type == VN_G_LAYER_POLYGON_CORNER_REAL)
	{
		if(do_build)
		{
			verse_send_g_polygon_set_corner_real64(node_id, id, 0,  0.1, 0.2, 0.3, 0.4);
			verse_send_g_polygon_set_corner_real64(node_id, id, 1,  0.2, 0.3, 0.4, 0.5);
			verse_send_g_polygon_set_corner_real64(node_id, id, 2,  0.6, 0.7, 0.8, 0.9);
			verse_send_g_polygon_set_corner_real64(node_id, id, 3,  1.0, 1.1, 1.2, 1.3);
			verse_send_g_crease_set_edge(node_id, "p_crease", 0);
		}
	}
	else if(type == VN_G_LAYER_POLYGON_FACE_UINT8)
	{
		if(do_build)
		{
			verse_send_g_polygon_set_face_uint8(node_id, id, 0,  0);
			verse_send_g_polygon_set_face_uint8(node_id, id, 1,  1);
			verse_send_g_polygon_set_face_uint8(node_id, id, 2,  2);
			verse_send_g_polygon_set_face_uint8(node_id, id, 3,  3);
		}
	}
	else if(type == VN_G_LAYER_POLYGON_FACE_UINT32)
	{
		if(do_build)
		{
			verse_send_g_polygon_set_face_uint32(node_id, id, 0,  0);
			verse_send_g_polygon_set_face_uint32(node_id, id, 1,  1);
			verse_send_g_polygon_set_face_uint32(node_id, id, 2,  2);
			verse_send_g_polygon_set_face_uint32(node_id, id, 3,  3);
		}
	}
	else if(type == VN_G_LAYER_POLYGON_FACE_REAL)
	{
		if(do_build)
		{
			verse_send_g_polygon_set_face_real64(node_id, id, 0, 0.1);
			verse_send_g_polygon_set_face_real64(node_id, id, 1, 0.2);
			verse_send_g_polygon_set_face_real64(node_id, id, 2, 0.3);
			verse_send_g_polygon_set_face_real64(node_id, id, 3, 0.4);
		}
	}
}

static void cb_g_crease_set_vertex(void *user_data, VNodeID node, const char *layer, uint32 def_crease)
{
	printf("Vertex crease set to '%s', with default of %u\n", layer, def_crease);
}

static void cb_g_crease_set_edge(void *user_data, VNodeID node, const char *layer, uint32 def_crease)
{
	printf("Edge crease set to '%s', with default of %u\n", layer, def_crease);
}

/* ------------------------------------------------------------------------------------------------------------------------- */

static void cb_connect_accept(void *user, VNodeID avatar, const char *address)
{
	printf("Connected to server at %s, as avatar %u.\n", address, avatar);
	if(do_build)
		printf(" Creating initial node, letting callbacks do the rest.\n");
	else
		printf(" Asking for all geometry nodes.\n");

	the_avatar = avatar;

	verse_send_node_index_subscribe(1 << V_NT_GEOMETRY);
	if(do_build)
		verse_send_node_create(0UL, V_NT_GEOMETRY, avatar);
}

static void cb_node_create(void *user, VNodeID node_id, VNodeType type, VNodeID owner_id)
{
	printf("hello\n");
	if(type == V_NT_GEOMETRY)
	{
		printf("Node %u has type %u, subcribing\n", node_id, type);
		verse_send_node_subscribe(node_id);
		if(owner_id == the_avatar)
		{
			printf("Creating XYZ layer\n");
			verse_send_g_layer_create(node_id, -1, "v_base", VN_G_LAYER_VERTEX_XYZ, 0, 0.0);

			printf("Creating Integer layer\n");
			verse_send_g_layer_create(node_id, -2, "v_skeleton", VN_G_LAYER_VERTEX_UINT32, 0, 0.0);

			printf("Creating Real layer\n");
			verse_send_g_layer_create(node_id, -3, "v_weight", VN_G_LAYER_VERTEX_REAL, 0, 0.0);
		}
	}
}

int main(int argc, char *argv[])
{
	const char	*server_address = "localhost";
	VSession	*s;

	if(argc == 2)
	{
		do_build = TRUE;
		printf("Argument detected, will build geometry\n");
	}
	else
		printf("No argument, will passively subscribe to geometry\n");

	if((s = verse_send_connect("geo-test", "password", server_address, NULL)) != NULL)
	{
		verse_callback_set((void *) verse_send_connect_accept,		(void *) cb_connect_accept, NULL);
		verse_callback_set((void *) verse_send_node_create,		(void *) cb_node_create,		NULL);

		verse_callback_set((void *) verse_send_g_layer_create,		(void *) cb_g_layer_create,		NULL);

		verse_callback_set((void *) verse_send_g_vertex_set_xyz_real64,	(void *) cb_g_vertex_set_xyz,		NULL);
		verse_callback_set((void *) verse_send_g_vertex_set_uint32,	(void *) cb_g_vertex_set_uint32,	NULL);
		verse_callback_set((void *) verse_send_g_vertex_set_real64,	(void *) cb_g_vertex_set_real64,	NULL);

		verse_callback_set((void *) verse_send_g_polygon_set_corner_uint32, (void *) cb_g_polygon_set_corner_uint32, NULL);
		verse_callback_set((void *) verse_send_g_polygon_set_corner_real64, (void *) cb_g_polygon_set_corner_real64, NULL);
		verse_callback_set((void *) verse_send_g_polygon_set_face_uint8,    (void *) cb_g_polygon_set_face_uint8,    NULL);
		verse_callback_set((void *) verse_send_g_polygon_set_face_uint32,   (void *) cb_g_polygon_set_face_uint32,   NULL);
		verse_callback_set((void *) verse_send_g_polygon_set_face_real64,   (void *) cb_g_polygon_set_face_real64,   NULL);

		verse_callback_set((void *) verse_send_g_crease_set_vertex,	(void *) cb_g_crease_set_vertex,	NULL);
		verse_callback_set((void *) verse_send_g_crease_set_edge,	(void *) cb_g_crease_set_edge,		NULL);

		while(1)
		{
			verse_callback_update(15000);
		}
		printf("Done, terminating\n");
		verse_session_destroy(s);
	}
	return 0;
}
