/*
 * A simple example program that just creates a number of nodes, and waits for the
 * server to respond with the assigned node IDs. Written by Emil Brink.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "verse.h"

static void cb_connect_accept(void *user, VNodeID avatar, const char *address, const uint8 *host_id)
{
	int	num = *(int *) user, i;

	printf("Connected, sending %d node_creates\n", num);
	verse_send_node_index_subscribe(~0);
	for(i = 0; i < num; i++)
		verse_send_node_create(~0, rand() % V_NT_NUM_TYPES, VN_OWNER_MINE);
}

static void cb_node_create(void *user, VNodeID node_id, VNodeType type, VNodeOwner owner)
{
	if(owner == VN_OWNER_MINE)
	{
		const char	*tn[] = { "object", "geometry", "material", "bitmap", "text", "curve", "audio" };
		int		*to_go = user;

		(*to_go)--;
		printf("Got node %u, type %d (%s), %d to go\n", node_id, type, tn[type], *to_go);
	}
}

int main(int argc, char *argv[])
{
	const char	*server = "localhost";
	int		i, to_go = 10;

	for(i = 1; argv[i]; i++)
	{
		if(strncmp(argv[i], "-ip=", 4) == 0)
			server = argv[i] + 4;
		else if(strncmp(argv[i], "-num=", 5) == 0)
			to_go = (int) strtol(argv[i] + 5, NULL, 0);
		else
			fprintf(stderr, "Ignoring unknown option '%s'\n", argv[i]);
	}

	if(to_go <= 0)
		return EXIT_SUCCESS;

	verse_callback_set((void *) verse_send_connect_accept,	(void *) cb_connect_accept,	&to_go);
	verse_callback_set((void *) verse_send_node_create,	(void *) cb_node_create,	&to_go);

	verse_send_connect("nodecreator", "<secret>", server, NULL);

	while(to_go > 0)
		verse_callback_update(10000);
	verse_send_connect_terminate(server, "All done, bye");

	return EXIT_SUCCESS;
}
