/*
** A first go at actually connecting to a Verse server, downloading an object, and rendering
** it to the screen using OpenGL. Fantastic stuff, no?
**
** 1999-10-29 -	Adapted for the new geometry node. As expected, this lead to cleaner code. Nice.
** 1999-11-18 -	Added more Quake-ish controls, at least now it's possible to strafe using the keyboard.
**		Still have the rather obnoxious mouse-look interface, though. We'll see how long it lasts.
** 2000-02-03 -	Now "supports" the new material system, meaning that it can under some circumstances
**		display textured polygons. *Please* note that this renderer is very far from being
**		serious, it's more like a prototype. The "support" for the material system implemented
**		in here would be laughable, if it wasn't so horrible. Basically, it will use the two
**		last seen Texture and Mapping fragments of a material, and simply assume that they are
**		used to provide color. Blenders, Outputs, and *all other fragments* are ignored!!
** 2000-08-30 -	Retargeted materials ugliness to handle the newly rewritten material system. Still
**		very ugly, tnough. It now assumes that a Texture fragment points at a Geometry in
**		its mapping field.
** 2000-09-03 -	Started to implement support for the new extrapolatable transform system. Neato.
** 2000-09-13 -	Cut out transform code, using new shared module.
** 2004-10-15 -	Ported to new Verse2 API, dropping much functionaity to get it up quickly. Time goes on.
*/

#define	__USE_BSD
#define _USE_MATH_DEFINES	/* Gets us M_PI from <math.h>. */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

#if defined _WIN32
#include <windows.h>
#define sinf	sin
#define	cosf	cos
#define sqrtf	sqrt
#define acosf	acos
#define snprintf	_snprintf
#endif

#if defined __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "verse.h"

#define	DEG2RAD(d)	((M_PI * (d)) / 180.0)
#define	RAD2DEG(d)	(((d) * 180.0) / M_PI)

/* ------------------------------------------------------------------------------------------------------------------------- */

#define	LAYER_NAME_SIZE			(16)		/* This *should* be the same as in the server, but doesn't have to be. */

/* ------------------------------------------------------------------------------------------------------------------------- */

#define	MENU_MASK			(0xFFFF0000)
#define	MENU_CAMERA			(1 << 16)
#define	MENU_CAMERA_FREEFLIGHT		(0)
#define	MENU_CAMERA_ISOMETRIC		(1)
#define	MENU_CAMERA_SATELLITE		(2)
#define	MENU_CAMERA_OBJECT		(3)

#define	MOVE_DELTA			0.5

/* ------------------------------------------------------------------------------------------------------------------------- */

/* FIXME: This is no longer used, but can potentially be re-introduced in update_bitmaps().
*/
/*#define	USE_GL_TEXSUBIMAGE2D*/			/* If this is on, we use a quick glTexSubImage2D() call to update the texture. */

typedef struct _NVLayer	NVLayer;
typedef struct _NPLayer	NPLayer;
typedef struct _NObject	NObject;

typedef struct {
	VNodeID		id;
	VNodeType	type;
	char		name[512];
} NHead;

/* Bitmap representation. Only 2D bitmaps are supported; 3D bitmaps (i.e. bitmaps with depth != 1) are ignored. */
typedef struct {
	NHead		head;
	uint16		width, height;

	GLuint		gl_id;			/* OpenGL texture object name. */

	VLayerID	l_red, l_green, l_blue;	/* The three "color_r", "color_g" and "color_b" layer IDs. */

	uint8		*pixel;			/* Local copy of entire bitmap in RGB format. */
	uint8		*tilebuf;		/* Temporary buffer holding one RGB tile. Copied from pixel. */
	uint32		dirty;			/* Increases when a tile changes. A bit gets set in dirtymap, and the node stored in dirty_bitmaps list. */
	uint8		*dirtymap;		/* One bit per tile, set when tile changes. */
	size_t		dirtymodulo;		/* Bytes per row of the bit matrix (dirtymap). */
} NBitmap;

typedef struct {
	NHead		head;
	char		map_u[LAYER_NAME_SIZE];
	char		map_v[LAYER_NAME_SIZE];
	VNodeID		texture;		/* A texture used in this material. This render is simplistic. */
	GList		*users;			/* Objects that use this material. */
} NMaterial;

typedef struct {
	GLdouble	x, y, z;
} VLDataXYZ;

typedef struct {
	uint32		v0, v1, v2, v3;
} PLDataCInt;

typedef struct {
	real64		v0, v1, v2, v3;
} PLDataCReal;

struct _NVLayer {
	VLayerID	id;
	char		name[LAYER_NAME_SIZE];
	VNGLayerType	type;
	uint32		def_integer;
	real64		def_real;
	union {
	VLDataXYZ	*xyz;
	uint32		*integer;
	real64		*real;
	} data;
};

struct _NPLayer {
	VLayerID	id;
	char		name[LAYER_NAME_SIZE];
	VNGLayerType	type;
	int32		def_integer;
	real64		def_real;
	union {
	PLDataCInt	*corner_int;
	PLDataCReal	*corner_real;
	uint32		*face_int;
	real64		*face_real;
	} data;
};

typedef struct NGBone	NGBone;

struct NGBone {
	char	weight[16];
	char	reference[16];
	uint16	parent_id;
	real64	pos[3];
	char	pos_curve[16];
	real64	rot[4];
	char	rot_curve[16];
	NGBone	*parent, *child[8];
	uint16	child_count;
};

typedef struct {
	NHead		head;
	uint32		num_vertices;
	GList		*vertex_layers;
	uint32		num_polygons;
	GList		*polygon_layers;
	NGBone		*bone;
	size_t		bone_alloc;
} NGeometry;

typedef void VTransformInfo;

typedef struct {
	uint16		id;
	VNodeID		node;
	uint32		target;
} ChildLink;

struct _NObject {
	NHead		head;
	real32		pos[4];			/* 4d for light. */
	VNQuat32	rot;
	real32		sx, sy, sz;		/* Scale. */
	real32		light[4];		/* 4d for glLightfv(). */
	NGeometry	*geometry;
	NMaterial	*material;
	gboolean	child;
	gboolean	visited;
	gboolean	hidden;
	GSList		*children;

	NPLayer		*map_u, *map_v;
};

typedef union {
	NHead		head;
	NObject		object;
	NGeometry	geometry;
	NMaterial	material;
} Node;

typedef struct {
	gboolean	enabled;
	size_t		sum, count;		/* For averaging. */
	size_t		history[32];		/* Averages stored here. */
} BufferHistory;

typedef struct {
	int		window;
	int		win_width, win_height;
	int		last_x, last_y;
	int		menu;
	int		obj_menu, cam_menu;
	boolean		visible;

	real32		home_pos[3], home_yaw, home_pitch;

	boolean		home_avatar;
	boolean		move_avatar;
	boolean		loud_keys;
	boolean		mouse_modify;
	int		mouse_button;

	boolean		show_time;
	boolean		show_camera;
	boolean		show_vertices;
	boolean		show_polygons;
	boolean		show_uvs;

	VNodeID		current;

	int		cam_mode;
	real32		cam_pos[3], cam_yaw, cam_pitch, cam_fwd[3];
	boolean		cam_active;

	boolean		move_active;
	real32		move_speed;

	boolean		object_cubes_active;
	boolean		wireframe_active;
	boolean		texture_active;
	boolean		axis_active;
	boolean		ground_active;
	boolean		lights_active;
	boolean		culling_active;

	int		start_time;
	int		frames;

	real32		yaw;

	int		timer_period;		/* Cap on max frame rate. In milliseconds (a GLUT limitation). */
	const char	*server;
	VSession	*session;		/* Connection to Verse server. */
	VNodeID		avatar;
	GHashTable	*nodes;
	GList		*objects;
	GList		*dirty_bitmaps;		/* Bitmaps that have changed, and need to be updated in GL. */

	BufferHistory	buf_hist;
	uint32		polygon_count;

	boolean		ping_active;
/*	GTimeVal	ping;*/
} GlobalData;

static GlobalData	global_data;

static void	g_polygon_layers_resize(NGeometry *geo, uint32 size);

/* ------------------------------------------------------------------------------------------------------------------------- */

static void r_strncpy(char *dst, size_t dst_max, const char *src)
{
	if(dst == NULL)
		return;
	if(dst_max < 1)
		return;
	if(src == NULL)
	{
		*dst = '\0';
		return;
	}
	dst_max--;
	for(; *src && dst_max; dst_max--)
		*dst++ = *src++;
	*dst = '\0';
}

static void ndb_init(void)
{
	GlobalData	*gd = &global_data;

	gd->nodes = g_hash_table_new(g_direct_hash, g_direct_equal);
	gd->objects = NULL;
	gd->dirty_bitmaps = NULL;
}

static gpointer ndb_lookup(VNodeID id)
{
	GlobalData	*gd = &global_data;
	Node		*node;

	if((node = g_hash_table_lookup(gd->nodes, GUINT_TO_POINTER(id))) != NULL)
		return node;
	return NULL;
}

static void ndb_build_glut_menu(void)
{
	gint	i, n;
	char	buf[128];
	GList	*iter;
	NObject	*obj;

	glutSetMenu(global_data.obj_menu);

	n = glutGet(GLUT_MENU_NUM_ITEMS);
	for(i = 1; i <= n; i++)
		glutRemoveMenuItem(1);
	glutAddMenuEntry("<None>", -1);
	for(iter = global_data.objects; iter != NULL; iter = g_list_next(iter))
	{
		obj = iter->data;
		g_snprintf(buf, sizeof buf, "%u %s", obj->head.id, obj->head.name);
		glutAddMenuEntry(buf, (gint) obj->head.id);
	}
}

static void ndb_remove(Node *node)
{
	g_hash_table_remove(global_data.nodes, node);
	if(node->head.type == V_NT_OBJECT)
		global_data.objects = g_list_remove(global_data.objects, node);
}

static void ndb_o_children_compute(void)
{
	GList	*iter;

	for(iter = global_data.objects; iter != NULL; iter = g_list_next(iter))
		((NObject *) iter->data)->child = FALSE;
	for(iter = global_data.objects; iter != NULL; iter = g_list_next(iter))
	{
		NObject	*obj = iter->data;
		GSList	*citer;

		for(citer = obj->children; citer != NULL; citer = g_slist_next(citer))
		{
			ChildLink	*ci = citer->data;
			NObject		*child;

			if((child = g_hash_table_lookup(global_data.nodes, GINT_TO_POINTER(ci->node))) != NULL)
				child->child = TRUE;
		}

	}
}

static gpointer ndb_o_new(VNodeID id)
{
	GlobalData	*gd = &global_data;
	NObject		*obj;

	if((obj = ndb_lookup(id)) != NULL)
		return obj;

	obj = g_malloc(sizeof *obj);
	obj->head.id = id;
	obj->head.type = V_NT_OBJECT;
	g_snprintf(obj->head.name, sizeof obj->head.name, "(unnamed)");

	memset(&obj->pos, 0, sizeof obj->pos);
	obj->pos[3] = 1.0f;
	memset(&obj->rot, 0, sizeof obj->rot);
	obj->sx = obj->sy = obj->sz = 1.0f;
	obj->light[0] = obj->light[1] = obj->light[2] = 0.0;
	obj->light[3] = 1.0f;
	obj->geometry = NULL;
	obj->material = NULL;
	obj->children = NULL;
	obj->child = FALSE;
	obj->visited = FALSE;
	obj->hidden = FALSE;

	obj->map_u = obj->map_v = NULL;

	g_hash_table_insert(gd->nodes, GUINT_TO_POINTER(id), obj);
	gd->objects = g_list_append(gd->objects, obj);

	ndb_o_children_compute();

	return obj;
}

static void quaternion_rotate(const VNQuat32 *rot);

/* Extract the forward vector from an object. This is done in a rather ... involved fashion, by
 * letting OpenGL compute the matrix and then reading it back. This is probably not very efficient,
 * but on the other hand we do it for at most one (1) object per frame, which should be tolerable.
*/
static void ndb_o_get_forward(const NObject *obj, real32 *fwd)
{
	GLfloat	mtx[16];

	glPushMatrix();
	glLoadIdentity();
	quaternion_rotate(&obj->rot);
	glGetFloatv(GL_MODELVIEW_MATRIX, mtx);
	fwd[0] = mtx[4];
	fwd[1] = mtx[5];
	fwd[2] = mtx[6];
	glPopMatrix();
}

static gpointer ndb_g_new(VNodeID id)
{
	GlobalData	*gd = &global_data;
	NGeometry	*geo;

	geo = g_malloc(sizeof *geo);
	geo->head.id	    = id;
	geo->head.type	    = V_NT_GEOMETRY;
	geo->num_vertices   = 0;
	geo->vertex_layers  = NULL;
	geo->num_polygons   = 0;
	geo->polygon_layers = NULL;
	geo->bone           = NULL;
	geo->bone_alloc     = 0u;

	g_hash_table_insert(gd->nodes, GUINT_TO_POINTER(id), geo);

	return geo;
}

static gpointer ndb_m_new(VNodeID id)
{
	GlobalData	*gd = &global_data;
	NMaterial	*mat;

	mat = g_malloc(sizeof *mat);

	mat->head.id	= id;
	mat->head.type	= V_NT_MATERIAL;

	mat->texture = 0UL;
	mat->users   = NULL;

	g_hash_table_insert(gd->nodes, GUINT_TO_POINTER(id), mat);

	return mat;
}

static gpointer ndb_b_new(VNodeID id)
{
	GlobalData	*gd = &global_data;
	NBitmap		*bit;

	bit = g_malloc(sizeof *bit);

	bit->head.id	= id;
	bit->head.type	= V_NT_BITMAP;

	bit->width = bit->height = 0u;
	bit->gl_id = 0;

	g_hash_table_insert(gd->nodes, GUINT_TO_POINTER(id), bit);

	return bit;
}


static void g_vertex_layer_create(NGeometry *geo, VLayerID l_id, const char *name, VNGLayerType type, uint32 def_integer, real64 def_real)
{
	NVLayer	*l;

	l = g_malloc(sizeof *l);
	l->id = l_id;
	g_snprintf(l->name, sizeof l->name, "%s", name);
	l->type = type;
	l->def_integer = def_integer;
	l->def_real    = def_real;

	switch(l->type)
	{
		case VN_G_LAYER_VERTEX_XYZ:
			l->data.xyz = NULL;
			break;
		case VN_G_LAYER_VERTEX_UINT32:
			l->data.integer = NULL;
			break;
		case VN_G_LAYER_VERTEX_REAL:
			l->data.real = NULL;
			break;
		default:
			;
	}
	geo->vertex_layers = g_list_append(geo->vertex_layers, l);
}

static void g_vertex_layers_resize(NGeometry *geo, uint32 size)
{
	GList	*iter;
	NVLayer	*l;

	for(iter = geo->vertex_layers; iter != NULL; iter = g_list_next(iter))
	{
		l = iter->data;
		switch(l->type)
		{
			case VN_G_LAYER_VERTEX_XYZ:
				l->data.xyz = g_realloc(l->data.xyz, size * sizeof *l->data.xyz);
				break;
			default:
				printf("Not resizing type %u-layer -- not implemented\n", l->type);
		}
	}
	geo->num_vertices = size;
}

static NVLayer * g_vertex_layer_find(NGeometry *geo, VLayerID l_id)
{
	GList	*iter;

	for(iter = geo->vertex_layers; iter != NULL; iter = g_list_next(iter))
	{
		if(((NVLayer *) iter->data)->id == l_id)
			return iter->data;
	}
	return NULL;
}

/*
static NVLayer * g_vertex_layer_find_by_name(NGeometry *geo, const char *name)
{
	GList	*iter;

	for(iter = geo->vertex_layers; iter != NULL; iter = g_list_next(iter))
	{
		if(strcmp(((NVLayer *) iter->data)->name, name) == 0)
			return iter->data;
	}
	return NULL;
}
*/

static void g_vertex_set_xyz(NGeometry *g, VLayerID l_id, uint32 v_id, real64 x, real64 y, real64 z)
{
	NVLayer	*l;

	if(v_id >= g->num_vertices)
		g_vertex_layers_resize(g, v_id + 1);

	if((l = g_vertex_layer_find(g, l_id)) != NULL)
	{
		l->data.xyz[v_id].x = x;
		l->data.xyz[v_id].y = y;
		l->data.xyz[v_id].z = z;
/*		printf("set vertex %u.%u.%u: (%g,%g,%g)\n", g->head.id, l_id, v_id,  x, y, z);*/
	}
}

static void g_polygon_layer_create(NGeometry *geo, VLayerID l_id, const char *name, VNGLayerType type, uint32 def_integer, uint32 def_real)
{
	NPLayer	*l;

	l = g_malloc(sizeof *l);
	l->id = l_id;
	g_snprintf(l->name, sizeof l->name, "%s", name);
	l->type = type;
	switch(type)
	{
		case VN_G_LAYER_POLYGON_CORNER_UINT32:
			l->data.corner_int = NULL;
			break;
		case VN_G_LAYER_POLYGON_CORNER_REAL:
			l->data.corner_real = NULL;
			break;
		case VN_G_LAYER_POLYGON_FACE_UINT32:
			l->data.face_int = NULL;
			break;
		case VN_G_LAYER_POLYGON_FACE_REAL:
			l->data.face_real = NULL;
			break;
		default:
			;
	}
	l->def_integer = def_integer;
	l->def_real = def_real;

	printf("polygon layer %u created, resizing layers\n", l_id);

	geo->polygon_layers = g_list_append(geo->polygon_layers, l);
	g_polygon_layers_resize(geo, geo->num_polygons);
}

static void g_polygon_layers_resize(NGeometry *geo, uint32 size)
{
	GList	*iter;
	NPLayer	*l;
	uint32	i;

	for(iter = geo->polygon_layers; iter != NULL; iter = g_list_next(iter))
	{
		l = iter->data;

		switch(l->type)
		{
			case VN_G_LAYER_POLYGON_CORNER_UINT32:
				l->data.corner_int = g_realloc(l->data.corner_int, size * sizeof *l->data.corner_int);
				for(i = geo->num_polygons; i < size; i++)
				{
					l->data.corner_int[i].v0 = l->data.corner_int[i].v1 =
					 l->data.corner_int[i].v2 = l->data.corner_int[i].v3 = UINT_MAX;
				}
				break;
			case VN_G_LAYER_POLYGON_CORNER_REAL:
				l->data.corner_real = g_realloc(l->data.corner_real, size * sizeof *l->data.corner_real);
				for(i = geo->num_polygons; i < size; i++)
				{
					l->data.corner_real[i].v0 = l->data.corner_real[i].v1 =
					l->data.corner_real[i].v2 = l->data.corner_real[i].v3 = l->def_real;
				}
				break;
			default:
				fprintf(stderr, "Don't know how to resize polygon layer of type %u\n", l->type);
		}
	}
	geo->num_polygons = size;
}

static NPLayer * g_polygon_layer_find(NGeometry *geo, VLayerID id)
{
	GList	*iter;

	for(iter = geo->polygon_layers; iter != NULL; iter = g_list_next(iter))
	{
		if(((NPLayer *) iter->data)->id == id)
			return iter->data;
	}
	return NULL;
}

static NPLayer * g_polygon_layer_find_by_name(NGeometry *geo, const char *name)
{
	GList	*iter;

	for(iter = geo->polygon_layers; iter != NULL; iter = g_list_next(iter))
	{
		if(strcmp(((NPLayer *) iter->data)->name, name) == 0)
			return iter->data;
	}
	return NULL;
}

static void g_polygon_set_corner_uint32(NGeometry *geo, VLayerID layer_id, uint32 poly_id, uint32 v0, uint32 v1, uint32 v2, uint32 v3)
{
	NPLayer	*l;

	if(poly_id >= geo->num_polygons)
		g_polygon_layers_resize(geo, poly_id + 1);

	if((l = g_polygon_layer_find(geo, layer_id)) != NULL)
	{
/*		printf("set corner ints %u.%u.%u: %u %u %u %u\n", geo->head.id, layer_id, poly_id, v0, v1, v2, v3);*/
		l->data.corner_int[poly_id].v0 = v0;
		l->data.corner_int[poly_id].v1 = v1;
		l->data.corner_int[poly_id].v2 = v2;
		l->data.corner_int[poly_id].v3 = v3;
	}
}

static void g_polygon_set_corner_real(NGeometry *geo, VLayerID layer_id, uint32 poly_id, real64 v0, real64 v1, real64 v2, real64 v3)
{
	NPLayer	*l;

	if(poly_id >= geo->num_polygons)
		g_polygon_layers_resize(geo, poly_id + 1);

	if((l = g_polygon_layer_find(geo, layer_id)) != NULL)
	{
		l->data.corner_real[poly_id].v0 = v0;
		l->data.corner_real[poly_id].v1 = v1;
		l->data.corner_real[poly_id].v2 = v2;
		l->data.corner_real[poly_id].v3 = v3;
/*		printf("Polygon %u.%u corner reals set to %g,%g,%g,%g\n", layer_id, poly_id, v0, v1, v2, v3);*/
	}
}

/* ------------------------------------------------------------------------------------------------------------------------- */

static void render_string(const char *string, GLfloat x, GLfloat y, GLfloat z)
{
	int	i;

	glRasterPos3f(x, y, z);
	for(i = 0; string[i]; i++)
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, string[i]);
}

static void render_number(unsigned int number, GLfloat x, GLfloat y, GLfloat z)
{
	char	buf[32];

	g_snprintf(buf, sizeof buf, "%u", number);
	render_string(buf, x, y, z);
}

static void quaternion_normalize(VNQuat32 *quat)
{
	float	qs = (quat->x * quat->x + quat->y * quat->y + quat->z * quat->z + quat->w * quat->w);

	if(fabs(qs - 1.0f) <= 1e-5f)	/* Good enough? Then skip the heavy lifting. */
		return;
	qs = 1.0f / sqrt(qs);
	quat->x *= qs;
	quat->y *= qs;
	quat->z *= qs;
	quat->w *= qs;
}

static void quaternion_rotate(const VNQuat32 *rot)
{
	float	cos_a = rot->w, sin_a_inv, angle;

	angle = 2.0f * acosf(cos_a);
	sin_a_inv = sqrtf(1.0f - cos_a * cos_a);
	if(fabs(sin_a_inv) < 0.0005f)
		sin_a_inv = 1.0f;
	sin_a_inv = 1.0f / sin_a_inv;
	glRotatef(RAD2DEG(angle), rot->x * sin_a_inv, rot->y * sin_a_inv, rot->z * sin_a_inv);
}

static void eulers_to_quat(VNQuat32 *quat, real32 heading, real32 attitude)
{
	real32	c1 = cosf(heading / 2.0f), c2 = cosf(attitude / 2.0f), c3 = 1.0f;
	real32	s1 = sinf(heading / 2.0f), s2 = sinf(attitude / 2.0f), s3 = 0.0f;

	quat->x = s1 * s2 * c3 + c1 * c2 * s3;
	quat->y = s1 * c2 * c3 + c1 * s2 * s3;
	quat->z = c1 * s2 * c3 - s1 * c2 * s3;
	quat->w = c1 * c2 * c3 - s1 * s2 * s3;
}

static void vtx_render(const NGeometry *node, const NVLayer *layer)
{
	unsigned int	i;
	const VLDataXYZ	*xyz;
	
	glColor3f(1.0, 0.1f, 0.3f);
	glPointSize(2.5f);
	glBegin(GL_POINTS);
	for(i = 0, xyz = layer->data.xyz; i < node->num_vertices; i++, xyz++)
		glVertex3f(xyz->x, xyz->y, xyz->z);
	glEnd();
/*	glDisable(GL_DEPTH_TEST);*/
	for(i = 0, xyz = layer->data.xyz; i < node->num_vertices; i++, xyz++)
	{
		render_number(i, xyz->x, xyz->y, xyz->z);
	}
/*	glEnable(GL_DEPTH_TEST);*/
}

static void poly_midpoint(real64 *mp, const VLDataXYZ *xyz, const PLDataCInt *pdef, unsigned int i, real32 out)
{
	mp[0] = xyz[pdef[i].v0].x + xyz[pdef[i].v1].x + xyz[pdef[i].v2].x;
	mp[1] = xyz[pdef[i].v0].y + xyz[pdef[i].v1].y + xyz[pdef[i].v2].y;
	mp[2] = xyz[pdef[i].v0].z + xyz[pdef[i].v1].z + xyz[pdef[i].v2].z;
	if(pdef[i].v3 != ~0u)
	{
		mp[0] += xyz[pdef[i].v3].x;
		mp[1] += xyz[pdef[i].v3].y;
		mp[2] += xyz[pdef[i].v3].z;
		mp[0] /= 4.0;
		mp[1] /= 4.0;
		mp[2] /= 4.0;
	}
	else
	{
		mp[0] /= 3.0;
		mp[1] /= 3.0;
		mp[2] /= 3.0;
	}
	if(out >= 0.0)	/* If displacement requested, compute normal and add it to the result. */
	{
		real64	e0[3], e1[3], n[3], f;

		e0[0] = xyz[pdef[i].v1].x - xyz[pdef[i].v0].x;
		e0[1] = xyz[pdef[i].v1].y - xyz[pdef[i].v0].y;
		e0[2] = xyz[pdef[i].v1].z - xyz[pdef[i].v0].z;

		e1[0] = xyz[pdef[i].v1].x - xyz[pdef[i].v2].x;
		e1[1] = xyz[pdef[i].v1].y - xyz[pdef[i].v2].y;
		e1[2] = xyz[pdef[i].v1].z - xyz[pdef[i].v2].z;

		n[0] = e0[1] * e1[2] - e0[2] * e1[1];
		n[1] = e0[2] * e1[0] - e0[0] * e1[2];
		n[2] = e0[0] * e1[1] - e0[1] * e1[0];

		f = 1.0 / sqrt(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);

		mp[0] += f * out * n[0];
		mp[1] += f * out * n[1];
		mp[2] += f * out * n[2];
	}
}

static void poly_render(const NGeometry *node, const NVLayer *vlay, const NPLayer *play)
{
	unsigned int		i, p = 0;
	const VLDataXYZ		*xyz;
	const PLDataCInt	*pdef;
	real64			mp[4];

	pdef = play->data.corner_int;
	xyz  = vlay->data.xyz;

	glColor3f(0.8f, 0.4f, 0.9f);
	for(i = 0; i < node->num_polygons; i++)
	{
		if(pdef[i].v0 == ~0u)
			continue;
		poly_midpoint(mp, xyz, pdef, i, 0.05);
		render_number(i, mp[0], mp[1], mp[2]);
		p++;
	}
}

static void vertex_blend(real32 factor, real64 *out, const real64 *a, const real64 *b)
{
	real32	ifa = 1.0f - factor;

	out[0] = factor * a[0] + ifa * b[0];
	out[1] = factor * a[1] + ifa * b[1];
	out[2] = factor * a[2] + ifa * b[2];
}

static void uvmap_render(const NGeometry *node, const NVLayer *vtx, const NPLayer *poly, const NPLayer *map_u, const NPLayer *map_v)
{
	unsigned int		i;
	const VLDataXYZ		*xyz = vtx->data.xyz;
	const PLDataCInt	*pdef = poly->data.corner_int;
	const PLDataCReal	*udat = map_u->data.corner_real, *vdat = map_v->data.corner_real;
	real64			mp[3], tp[3];
	char			buf[16];

	glColor3f(0.2, 0.7f, 0.9f);
	for(i = 0; i < node->num_polygons; i++)
	{
		if(pdef[i].v0 == ~0u || pdef[i].v1 == ~0u || pdef[i].v2 == ~0u)
			continue;
		poly_midpoint(mp, xyz, pdef, i, 0.0);
		g_snprintf(buf, sizeof buf, "%g,%g", udat[i].v0, vdat[i].v0);
		vertex_blend(0.1, tp, mp, (real64 *) &xyz[pdef[i].v0]);
		render_string(buf, tp[0], tp[1], tp[2]);
		g_snprintf(buf, sizeof buf, "%g,%g", udat[i].v1, vdat[i].v1);
		vertex_blend(0.1, tp, mp, (real64 *) &xyz[pdef[i].v1]);
		render_string(buf, tp[0], tp[1], tp[2]);
		g_snprintf(buf, sizeof buf, "%g,%g", udat[i].v2, vdat[i].v2);
		vertex_blend(0.1, tp, mp, (real64 *) &xyz[pdef[i].v2]);
		render_string(buf, tp[0], tp[1], tp[2]);
		if(pdef[i].v3 == ~0u)
			continue;
		g_snprintf(buf, sizeof buf, "%g,%g", udat[i].v3, vdat[i].v3);
		vertex_blend(0.1, tp, mp, (real64 *) &xyz[pdef[i].v3]);
		render_string(buf, tp[0], tp[1], tp[2]);
	}
}

static void bone_render(const NGBone *bone, const NGeometry *geo)
{
/*	GLdouble	mtx[16];*/

	glPushMatrix();
	glTranslated(bone->pos[0], bone->pos[1], bone->pos[2]);
	glBegin(GL_POINTS);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glEnd();
	if(bone->child_count > 0)
	{
		register int	i;
		glPushMatrix();
		for(i = 0; i < bone->child_count; i++)
			bone_render(bone->child[i], geo);
		glPopMatrix();
	}
	glPopMatrix();
}

static void skeleton_render(const NObject *obj, const NGeometry *geo)
{
	unsigned int	i, rc = 0, j;
	NGBone		*bone, *root[32];

	/* Clear child pointers. */
	for(i = 0, bone = geo->bone; i < geo->bone_alloc; i++, bone++)
	{
		if(bone->weight[0] == '\0')
			continue;
		bone->child_count = 0;
	}
	/* Set up parent and child pointers. */
	for(i = 0, bone = geo->bone; i < geo->bone_alloc; i++, bone++)
	{
		if(bone->weight[0] == '\0')
			continue;
		if(bone->parent_id >= geo->bone_alloc || geo->bone[bone->parent_id].weight[0] == '\0')	/* Parent-less? */
		{
			if(rc == sizeof root / sizeof *root)
			{
				fprintf(stderr, "render: Too many root bones in hierarchy, max %u supported\n", sizeof root / sizeof *root);
				break;
			}
			root[rc++] = bone;
			bone->parent = NULL;
		}
		else
		{
			NGBone	*p = geo->bone + bone->parent_id;

			if(p->child_count == sizeof p->child / sizeof *p->child)
			{
				fprintf(stderr, "render: Bone %u has too many children, max %u supported\n", bone->parent_id, sizeof p->child / sizeof *p->child);
				break;
			}
			bone->parent = p;
			p->child[p->child_count++] = bone;
		}
	}
	/* At this point, I guess we should look for cycles. Rendering is more fun, though. */
	glPointSize(5.0);
	glColor3f(0.8f, 0.8f, 1.0f);
	glPushMatrix();
	for(i = 0; i < rc; i++)
	{
		bone = root[i];
		glTranslated(bone->pos[0], bone->pos[1], bone->pos[2]);
		for(j = 0; j < bone->child_count; j++)
			bone_render(bone->child[j], geo);
/*		bone->pos[0] += 0.1;
		if(bone->pos[0] > 4.0)
			bone->pos[0] = 0.0;
*/	}
	glPopMatrix();
	glPointSize(1.0);
}

static void normal_compute(const VLDataXYZ *xyz, uint32 v0, uint32 v1, uint32 v2)
{
	real64	v10[3], v12[3], n[3];

	v10[0] = xyz[v1].x - xyz[v0].x;
	v10[1] = xyz[v1].y - xyz[v0].y;
	v10[2] = xyz[v1].z - xyz[v0].z;

	v12[0] = xyz[v1].x - xyz[v2].x;
	v12[1] = xyz[v1].y - xyz[v2].y;
	v12[2] = xyz[v1].z - xyz[v2].z;

	n[0] = v10[1] * v12[2] - v10[2] * v12[1];
	n[1] = v10[2] * v12[0] - v10[0] * v12[2];
	n[2] = v10[0] * v12[1] - v10[1] * v12[0];

	glNormal3d(n[0], n[1], n[2]);
}

#define	VPRINT(xyz,pdef,v)	printf("Polygon %u vertex %s, pushing (%g,%g,%g) to OpenGL\n", i, #v, xyz[pdef->v].x, xyz[pdef->v].y, xyz[pdef->v].z);

static void object_render(NObject *obj)
{
	NGeometry	*geo = NULL;
	uint32		rgb;
	GSList		*iter;

	if(obj->visited || obj->hidden)
		return;
	obj->visited = TRUE;

	/* Set the transform. */
	glPushMatrix();
	glTranslated(obj->pos[0], obj->pos[1], obj->pos[2]);
	quaternion_rotate(&obj->rot);
	glScaled(obj->sx, obj->sy, obj->sz);

	geo = obj->geometry;
/*	if(geo != NULL)
	{
		printf("object at %p (%u) has geo at %p (%u)\n", obj, obj->head.id, obj->geometry, obj->geometry != NULL ? obj->geometry->head.id : ~0u);
		printf(" geometry has %u vertices, %u polys\n", geo->num_vertices, geo->num_polygons);
	}
*/
	if(geo == NULL && global_data.object_cubes_active)		/* No geometry? Then render something else. */
	{
		if(obj->light[0] != 0.0 || obj->light[1] != 0.0 || obj->light[2] != 0.0)
		{
			/* Render lightsource a bit less intrusively, as small wire spheres. */
			glColor3f(obj->light[0], obj->light[1], obj->light[2]);
			glutWireSphere(0.25, 4, 4);
		}
		else
		{
			/* Render other object nodes as wireframe cubes. */
			glColor3f(0.0, 0.375, 1.0);
			glutWireCube(2.0);
		}
	}
	else if(geo != NULL && geo->num_vertices > 0 && geo->num_polygons > 0)
	{
		NVLayer		*vtxlay;
		NPLayer		*play;
		PLDataCInt	*pdef;
		NBitmap		*bit = NULL;

		if(((vtxlay = g_vertex_layer_find(geo, 0)) != NULL) && ((play = g_polygon_layer_find(geo, 1)) != NULL))
		{
			const VLDataXYZ	*xyz = vtxlay->data.xyz;
			uint32	i, poly_skip = 0;

			if(obj->material != NULL)
				bit = ndb_lookup(obj->material->texture);

			obj->map_u = g_polygon_layer_find_by_name(geo, "map_u");
			obj->map_v = g_polygon_layer_find_by_name(geo, "map_v");

			if((obj->material == 0U) || (obj->map_u == NULL) || (obj->map_v == NULL) || (global_data.texture_active == FALSE))
			{
/*				glColor3f(1.0f, 1.0f, 1.0f);*/
				for(i = 0; i < geo->num_polygons; i++)
				{
					pdef = &play->data.corner_int[i];
					if(pdef->v0 > geo->num_vertices || pdef->v1 > geo->num_vertices || pdef->v2 > geo->num_vertices)
					{
						poly_skip++;
						continue;
					}
/*					printf(" %u: %g %g %g ", i, xyz[pdef->v0].x, xyz[pdef->v1].x, xyz[pdef->v2].x);
					if(pdef->v3 < UINT_MAX)
						printf("%g", xyz[pdef->v3].x);
					printf("\n");
*/					if(xyz[pdef->v0].x == V_REAL64_MAX || xyz[pdef->v1].x == V_REAL64_MAX ||
					   xyz[pdef->v2].x == V_REAL64_MAX ||
					   (pdef->v3 < geo->num_vertices && xyz[pdef->v3].x == V_REAL64_MAX))
					{
						continue;	/* Vertex is deleted. */
					}
/*					if(xyz[pdef->v0].x == xyz[pdef->v0].y &&
					   xyz[pdef->v0].y == xyz[pdef->v0].z)
						printf("vertex 0, %u is all %g, suspicious\n", pdef->v0, xyz[pdef->v0].x);
					if(xyz[pdef->v1].x == xyz[pdef->v1].y &&
					   xyz[pdef->v1].y == xyz[pdef->v1].z)
						printf("vertex 1, %u is all %g, suspicious\n", pdef->v1, xyz[pdef->v1].x);
					if(xyz[pdef->v2].x == xyz[pdef->v2].y &&
					   xyz[pdef->v2].y == xyz[pdef->v2].z)
						printf("vertex 2, %u is all %g, suspicious\n", pdef->v2, xyz[pdef->v2].x);
*/
					if(global_data.lights_active)
						glColor3f(0.7f, 0.7f, 0.8f);
					else
					{
						rgb = rand();
						glColor3ub((rgb >> 16) & 0xFF, (rgb >> 8) & 0xFF, rgb & 0xFF);
					}
/*					printf("%u: %u %u %u %u\n", i, pdef->v0, pdef->v1, pdef->v2, pdef->v3);*/
					glBegin(GL_TRIANGLE_FAN);
					if(global_data.lights_active)
						normal_compute(xyz, pdef->v0, pdef->v1, pdef->v2);
					glVertex3d(xyz[pdef->v0].x, xyz[pdef->v0].y, xyz[pdef->v0].z);
					glVertex3d(xyz[pdef->v1].x, xyz[pdef->v1].y, xyz[pdef->v1].z);
					glVertex3d(xyz[pdef->v2].x, xyz[pdef->v2].y, xyz[pdef->v2].z);
					if(pdef->v3 < geo->num_vertices)
						glVertex3d(xyz[pdef->v3].x, xyz[pdef->v3].y, xyz[pdef->v3].z);
					glEnd();
					global_data.polygon_count += 2 - (pdef->v3 != ~0u);
				}
				if(poly_skip > 0)
					printf("render: skipped %u broken polygons in object %u (%s), geo %u (%s)\n",
					       poly_skip, obj->head.id, obj->head.name, geo->head.id, geo->head.name);
			}
			else if(global_data.texture_active && (bit && obj->map_u && obj->map_v))
			{
				PLDataCReal	*tdefu, *tdefv;

				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, bit->gl_id);
				glColor3f(1.0f, 1.0f, 1.0f);
				for(i = 0; i < geo->num_polygons; i++)
				{
					pdef  = &play->data.corner_int[i];
					if(pdef->v0 == UINT_MAX)
						continue;
					tdefu = &obj->map_u->data.corner_real[i];
					tdefv = &obj->map_v->data.corner_real[i];
					glBegin(GL_TRIANGLE_FAN);
					glTexCoord2f(tdefu->v0, tdefv->v0);
					glVertex3d(xyz[pdef->v0].x, xyz[pdef->v0].y, xyz[pdef->v0].z); 
					glTexCoord2f(tdefu->v1, tdefv->v1);
					glVertex3d(xyz[pdef->v1].x, xyz[pdef->v1].y, xyz[pdef->v1].z);
					glTexCoord2f(tdefu->v2, tdefv->v2);
					glVertex3d(xyz[pdef->v2].x, xyz[pdef->v2].y, xyz[pdef->v2].z); 
					if(pdef->v3 != UINT_MAX)
					{
						glTexCoord2f(tdefu->v3, tdefv->v3);
						glVertex3d(xyz[pdef->v3].x, xyz[pdef->v3].y, xyz[pdef->v3].z);
					}
					glEnd();
				}
				glDisable(GL_TEXTURE_2D);
			}
			if(global_data.show_vertices)
				vtx_render(geo, vtxlay);
			if(global_data.show_polygons)
				poly_render(geo, vtxlay, play);
			if(global_data.show_uvs && obj->map_u != NULL && obj->map_v != NULL)
				uvmap_render(geo, vtxlay, play, obj->map_u, obj->map_v);
		}
	}
	if(geo != NULL && geo->bone_alloc > 0)
		skeleton_render(obj, geo);
	/* Before popping the transform, iterate over all child nodes and render them, too. */
	for(iter = obj->children; iter != NULL; iter = g_slist_next(iter))
	{
		const ChildLink	*ci = iter->data;
		NObject		*ch;

		if((ch = g_hash_table_lookup(global_data.nodes, GINT_TO_POINTER(ci->node))) != NULL && ch->head.type == V_NT_OBJECT)
			object_render(ch);
		else
			printf("failed to render child node %u (link %d)\n", ci->node, ci->id);
	}
	glPopMatrix();
}

/* Go through internal object data base, and render all objects found. */
static void ndb_render(void)
{
	GList		*iter;
	NObject		*obj;
	uint32		i, toplevels = 0;

	if(global_data.wireframe_active)
		glPolygonMode(GL_FRONT, GL_LINE);
	if(global_data.culling_active)
		glEnable(GL_CULL_FACE);
	else
		glDisable(GL_CULL_FACE);

	global_data.polygon_count = 0;

	for(iter = global_data.objects; iter != NULL; iter = g_list_next(iter))
		((NObject *) iter->data)->visited = FALSE;
	for(iter = global_data.objects; iter != NULL; iter = g_list_next(iter))
	{
		NObject	*obj = iter->data;

		if(!obj->child)
		{
			srand((uint32) obj);	/* Set random seed, for each object (for hide). */
			object_render(obj);
			toplevels++;
		}
	}
	for(iter = global_data.objects; iter != NULL; iter = g_list_next(iter))
	{
		obj = iter->data;
		if(obj->geometry == NULL)
			continue;
		if(obj->head.id == global_data.current)
			glColor3ub(0xFF, 0xFF, 0x00);
		else
			glColor3ub(0x00, 0x80, 0xFF);
		glPushMatrix();
		glTranslated(obj->pos[0], obj->pos[1], obj->pos[2]);
		glRasterPos3f(0.0f, 0.0f, 0.0f);
		for(i = 0; obj->head.name[i]; i++)
			glutBitmapCharacter(GLUT_BITMAP_8_BY_13, obj->head.name[i]);
		glPopMatrix();
	}
	if(global_data.show_time)
	{
		uint32	now_s, now_f;

		verse_session_get_time(&now_s, &now_f);
		glColor3f(0.7, 0.9, 0.8);
		render_number(now_s, 0.0, 10.0, 0.0);
		render_number(now_f, 0.0, 9.7, 0.0);
	}
	if(global_data.show_camera)
	{
		printf("Camera at (%g,%g,%g), pitch=%g yaw=%g\n", global_data.cam_pos[0], global_data.cam_pos[1], global_data.cam_pos[2],
			global_data.cam_pitch, global_data.cam_yaw);
	}
	glPolygonMode(GL_FRONT, GL_FILL);
}

/* ------------------------------------------------------------------------------------------------------------------------- */

static void cb_ping(void *user, const char *server, const char *message)
{
	uint32	ns, nf, ts, tf;

	verse_session_get_time(&ns, &nf);
	if(sscanf(message, "%u.%u", &ts, &tf) == 2)
	{
		double	s = (ns - ts) + (nf - tf) / 4294967295.0;
		printf("Got ping response from \"%s\" in %gs\n", server, s);
	}
	else
		printf("Render: Couldn't parse ping\n");
}

static void cb_connect_accept(void *user, VNodeID avatar, const char *address)
{
	char	buf[32];

	printf("connected to server as %u\n", avatar);
	global_data.avatar = avatar;
	verse_send_node_index_subscribe(1 << V_NT_OBJECT);		/* Start walking world graph. */
	sprintf(buf, "simple-renderer-%u", (unsigned int) avatar);
	verse_send_node_name_set(avatar, buf);
}

static void cb_node_create(void *user_data, VNodeID node_id, VNodeType type, VNodeID owner_id)
{
	printf("Now in cb_node_create(), node=%lu type=%d owner=%lu\n", (unsigned long) node_id, type, (unsigned long) owner_id);

	if((type == V_NT_OBJECT) && (node_id != global_data.avatar))
	{
		ndb_o_new(node_id);

		verse_send_node_subscribe(node_id);
		verse_send_o_transform_subscribe(node_id, VN_FORMAT_REAL32);
	}
}

static void cb_node_name_set(void *user_data, VNodeID node_id, const char *name)
{
	NHead	*obj;

	if((obj = ndb_lookup(node_id)) != NULL)
	{
		if(name[0] == '\0')
			g_snprintf(obj->name, sizeof obj->name, "(Anonymous)");
		else
			g_snprintf(obj->name, sizeof obj->name, "%s", name);
		printf("name of %u set to '%s'\n", node_id, name);
		ndb_build_glut_menu();
	}	
}

static void cb_node_destroy(void *user, VNodeID node_id)
{
	Node	*node;

	if((node = ndb_lookup(node_id)) != NULL)
	{
		ndb_remove(node);
		/* FIXME: This might leak memory! */
		g_free(node);
		ndb_build_glut_menu();
		if(global_data.current == node_id)
			global_data.current = ~0u;
	}
}

#if 0

static void obj_set_material(NObject *obj, NMaterial *mat)
{
	obj->material = mat;
	obj->material->users = g_list_append(obj->material->users, obj);
	obj->map_u = g_polygon_layer_find_by_name(obj->geometry, mat->map_u);
	obj->map_v = g_polygon_layer_find_by_name(obj->geometry, mat->map_v);
}

#endif

static void cb_o_link_set(void *user_data, VNodeID node_id, uint16 link_id, VNodeID link, const char *label, uint32 target_id)
{
	NObject	*obj;

	printf("set link %u, from object %u to node %u, label '%s'\n", (int) link_id, node_id, link, label);
	if((obj = ndb_lookup(node_id)) != NULL)
	{
		if(obj->head.type != V_NT_OBJECT)
		{
			printf(" aborting, node %u is type %d, not %d\n", obj->head.id, obj->head.type, V_NT_OBJECT);
			return;
		}
		if(strcmp("geometry", label) == 0 /*&& target_id == 0*/)
		{
			NGeometry	*geo;

			if((geo = ndb_lookup(link)) == NULL)
			{
				geo = ndb_g_new(link);
				verse_send_node_subscribe(link);
			}
			else if(geo->head.type != V_NT_GEOMETRY)
				geo = NULL;
			obj->geometry = geo;
			if(geo != NULL)
				printf("geometry link set to %p (id %u)\n", geo, geo->head.id);
		}
		else if(strcmp("child", label) == 0)
		{
			ChildLink	*ci = malloc(sizeof *ci);
			ci->id = link_id;	/* Not very well handled, at the moment. */
			ci->node = link;
			ci->target = target_id;
			obj->children = g_slist_prepend(obj->children, ci);
			printf("need to store a child link, there are now %u\n", g_slist_length(obj->children));
			ndb_o_children_compute();
		}
		else
			printf("unknown link label '%s', target=%u\n", label, target_id);
/*		break;
		case V_NT_MATERIAL:
		{
			NMaterial	*mat;
			if((mat = ndb_lookup(id)) == NULL)
			{
				mat = ndb_m_new(id);
				vll_send_subscribe(id, type);
			}
			obj_set_material(obj, mat);
			break;
		}
		default:
			printf("Ignoring link of type %u\n", type);
		}
*/
	}
	else
		printf("nothing known of object %u, can't look it up\n", node_id);
}

static void cb_o_hide(void *user, VNodeID node_id, boolean hidden)
{
	NObject	*obj;

	if((obj = ndb_lookup(node_id)) != NULL)
		obj->hidden = hidden;
}

static void cb_o_transform_pos_real32(void *user, VNodeID node_id, uint32 time_s, uint32 time_f,
				      const real32 *pos, const real32 *speed, const real32 *accelerate, const real32 *drag_normal, real32 drag)
{
	NObject	*obj;

	if((obj = ndb_lookup(node_id)) != NULL)
	{
/*		printf("set position of object %u to (%g,%g,%g)\n", node_id, pos[0], pos[1], pos[2]);*/
		obj->pos[0] = pos[0];
		obj->pos[1] = pos[1];
		obj->pos[2] = pos[2];
	}
}

static void cb_o_transform_rot_real32(void *user, VNodeID node_id, uint32 time_s, uint32 time_f, const VNQuat32 *rot,
				      const VNQuat32 *speed, const VNQuat32 *accelerate, const VNQuat32 *drag_normal, real32 drag)
{
	NObject	*obj;

	if((obj = ndb_lookup(node_id)) != NULL)
	{
/*		printf("set rotation of object %u to (%g,%g,%g,%g)\n", node_id,
		       rot->x, rot->y, rot->z, rot->w);
*/		obj->rot = *rot;
		quaternion_normalize(&obj->rot);
	}
}

static void cb_o_transform_scale_real32(void *user, VNodeID node_id, real32 x, real32 y, real32 z)
{
	NObject	*obj;

	if((obj = ndb_lookup(node_id)) != NULL)
	{
		if(x == 0.0f || y == 0.0f || z == 0.0f)
			return;
/*		printf("set scale of object %u to (%g,%g,%g)\n", node_id, x, y, z);*/
		obj->sx = x;
		obj->sy = y;
		obj->sz = z;
	}
}
	
static void cb_o_light_set(void *user, VNodeID node_id, real64 red, real64 green, real64 blue)
{
	NObject	*obj;

	if((obj = ndb_lookup(node_id)) != NULL)
	{
		obj->light[0] = red;
		obj->light[1] = green;
		obj->light[2] = blue;
		printf("light of object %u set to (%g,%g,%g)\n", node_id, red, green, blue);
	}
}

#if 0

/* Ugly hack to "support" new material linking system. */
static void cb_o_link_material_set(void *user_data, VNodeID node_id, VNodeID node, uint16 nlut)
{
	NObject	*obj;

	if(nlut != 0U)
	{
		printf("Object %lu: Got O_LinkMaterialSet with NLUT=%u, which is non-zero\n", (unsigned long) node_id, nlut);
		return;
	}

	if((obj = ndb_lookup(node_id)) != NULL)
	{
		if(obj->head.type == V_NT_OBJECT)
		{
			NMaterial	*mat;

			printf("Object %lu: link type V_NT_MATERIAL set to %lu\n", (unsigned long) node_id, (unsigned long) node);
			if((mat = ndb_lookup(node)) == NULL)
			{
				mat = ndb_m_new(node);
				vll_send_subscribe(node, V_NT_MATERIAL);
			}
			obj_set_material(obj, mat);
		}
	}
}

/* ------------------------------------------------------------------------------------------------------------------------- */

#endif

static void cb_g_layer_create(void *user_data, VNodeID node_id, VLayerID layer_id, const char *name, VNGLayerType type, uint32 def_integer, real64 def_real)
{
	NGeometry	*geo;

	if((geo = ndb_lookup(node_id)) != NULL)
	{
		/* Filter out non-handled layer types. */
		switch(type)
		{
			case VN_G_LAYER_VERTEX_XYZ:		/* Fall-through! */
			case VN_G_LAYER_POLYGON_CORNER_UINT32:
			case VN_G_LAYER_POLYGON_CORNER_REAL:
				break;
			default:
				printf("Not subscribing to geometry layer '%s', type %u - not handled\n", name, type);
				return;
		}
		if(type < VN_G_LAYER_POLYGON_CORNER_UINT32)
			g_vertex_layer_create(geo, layer_id, name, type, def_integer, def_real);
		else
			g_polygon_layer_create(geo, layer_id, name, type, def_integer, def_real);
		printf("Subscribing to geometry layer %u, type %u (name '%s')\n", layer_id, type, name);
		verse_send_g_layer_subscribe(node_id, layer_id, VN_FORMAT_REAL64);
	}
}

static void cb_g_vertex_set_xyz_real64(void *user_data, VNodeID node_id, VLayerID layer_id, uint32 vertex_id, real64 x, real64 y, real64 z)
{
	NGeometry	*geo;

/*	printf("Setting vertex %u.%u.%u to (%g,%g,%g)\n", node_id, layer_id, vertex_id, x, y, z);*/
	if(((geo = ndb_lookup(node_id)) != NULL) && (geo->head.type == V_NT_GEOMETRY))
		g_vertex_set_xyz(geo, layer_id, vertex_id, x, y, z);
}

static void cb_g_vertex_set_xyz_real32(void *user_data, VNodeID node_id, VLayerID layer_id, uint32 vertex_id, real32 x, real32 y, real32 z)
{
	printf("Got vertex %u.%u in REAL32 format, very unexpected -- server bug?\n", node_id, vertex_id);
	cb_g_vertex_set_xyz_real64(user_data, node_id, layer_id, vertex_id, x, y, z);
}

static void cb_g_vertex_delete(void *user_data, VNodeID node_id, uint32 vertex_id)
{
	NGeometry	*geo;

	if((geo = ndb_lookup(node_id)) != NULL)
	{
		if(vertex_id < geo->num_vertices)
		{
			NVLayer	*base;

			if((base = g_vertex_layer_find(geo, 0)) != NULL)
			{
				VLDataXYZ	*vtx = &base->data.xyz[vertex_id];

				vtx->x = vtx->y = vtx->z = V_REAL64_MAX;
			}
		}
	}
}

static void cb_g_polygon_set_corner_uint32(void *user_data, VNodeID node_id, VLayerID layer_id, uint32 poly_id, uint32 v0, uint32 v1, uint32 v2, uint32 v3)
{
	NGeometry	*geo;

	if((geo = ndb_lookup(node_id)) != NULL)
	{
		g_polygon_set_corner_uint32(geo, layer_id, poly_id, v0, v1, v2, v3);
/*		printf("Polygon corners %u.%u.%u set to %u %u %u %u\n", node_id, layer_id, poly_id, v0, v1, v2, v3);*/
	}
}

static void cb_g_polygon_set_corner_real64(void *user_data, VNodeID node_id, VLayerID layer_id, uint32 poly_id, real64 v0, real64 v1, real64 v2, real64 v3)
{
	NGeometry	*geo;

	if((geo = ndb_lookup(node_id)) != NULL)
		g_polygon_set_corner_real(geo, layer_id, poly_id, v0, v1, v2, v3);
}

static void cb_g_polygon_delete(void *user_data, VNodeID node_id, uint32 polygon_id)
{ 
	NGeometry	*geo;

	if(((geo = ndb_lookup(node_id)) != NULL) && (polygon_id < geo->num_polygons))
	{
		NPLayer	*base;

		if((base = g_polygon_layer_find(geo, 1)) != NULL)
		{
			base->data.corner_int[polygon_id].v0 = UINT_MAX;
			base->data.corner_int[polygon_id].v1 = UINT_MAX;
			base->data.corner_int[polygon_id].v2 = UINT_MAX;
			base->data.corner_int[polygon_id].v3 = UINT_MAX;
		}
		else
			printf(" No base layer in node?\n");
	}
}

static void cb_g_bone_create(void *user_data, VNodeID node_id, uint16 bone_id, const char *weight, const char *reference, uint16 parent_id,
			     real64 pos_x, real64 pos_y, real64 pos_z, const char *pos_curve,
			     real64 rot_x, real64 rot_y, real64 rot_z, real64 rot_w, const char *rot_curve)
{
	NGeometry	*geo;
	NGBone		*bone;

	if((geo = ndb_lookup(node_id)) == NULL || geo->head.type != V_NT_GEOMETRY)
		return;

	printf("need to store bone %u, at (%g,%g,%g) from parent %u\n", bone_id, pos_x, pos_y, pos_z, parent_id);
	if(bone_id >= geo->bone_alloc)
	{
		size_t	na = bone_id + 8, i;
		NGBone	*nb;

		nb = realloc(geo->bone, na * sizeof *nb);
		if(nb == NULL)
		{
			fprintf(stderr, "render: Unable to store new bone with ID=%u, memory allocation failed\n", bone_id);
			return;
		}
		for(i = geo->bone_alloc; i < na; i++)
		{
			nb[i].weight[0] = '\0';
			printf("clearing bone %u\n", i);
		}
		geo->bone = nb;
		geo->bone_alloc = na;
	}
	bone = geo->bone + bone_id;
	r_strncpy(bone->weight, sizeof bone->weight, weight);
	r_strncpy(bone->reference, sizeof bone->reference, reference);
	bone->parent_id = parent_id;
	bone->parent = NULL;
	bone->pos[0] = pos_x;
	bone->pos[1] = pos_y;
	bone->pos[2] = pos_z;
	r_strncpy(bone->pos_curve, sizeof bone->pos_curve, pos_curve);
	bone->rot[0] = rot_x;
	bone->rot[1] = rot_y;
	bone->rot[2] = rot_z;
	bone->rot[3] = rot_w;
	r_strncpy(bone->rot_curve, sizeof bone->rot_curve, rot_curve);
}

/* ------------------------------------------------------------------------------------------------------------------------- */

static void cb_b_dimensions_set(void *user_data, VNodeID node_id, uint16 width, uint16 height, uint16 depth)
{
	NBitmap	*bit;

	printf("Bitmap node %u is %ux%ux%u\n", node_id, width, height, depth);
	if(depth != 1)
	{
		printf(" Bitmaps with depth != 1 are not supported, ignoring node %u\n", node_id);
		return;
	}
	if((bit = ndb_lookup(node_id)) != NULL)
	{
		uint32	x, y;
		uint8	*ptr;

		bit->width  = width;
		bit->height = height;

		printf(" Allocating %u bytes for a %ux%u RGB image\n", 3 * bit->width * bit->height, width, height);
		bit->pixel = g_malloc(3 * bit->width * bit->height);
		for(y = 0; y < bit->height; y++)
		{
			ptr = bit->pixel + 3 * y * bit->width;
			for(x = 0; x < bit->width; x++, ptr += 3)
			{
				*ptr = ((y >> 3) & 1)  ?
					((x >> 3) & 1) ? 0x00 : 0xFF :
					((x >> 3) & 1) ? 0xFF : 0x00;
				*(ptr + 1) = *(ptr + 2) = 0x00;
			}
		}
		glGenTextures(1, &bit->gl_id);
		glBindTexture(GL_TEXTURE_2D, bit->gl_id);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, bit->width, bit->height, 0, GL_RGB, GL_UNSIGNED_BYTE, bit->pixel);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		bit->tilebuf = g_malloc(3 * VN_B_TILE_SIZE * VN_B_TILE_SIZE);
		bit->dirtymodulo = ((bit->width / VN_B_TILE_SIZE) + CHAR_BIT - 1) / CHAR_BIT;
		bit->dirtymap = g_malloc(bit->dirtymodulo * (bit->height / VN_B_TILE_SIZE));
		memset(bit->dirtymap, 0, bit->dirtymodulo * (bit->height / VN_B_TILE_SIZE));
		bit->dirty = FALSE;
	}
	else
		fprintf(stderr, "RENDER: Got dimensions_set for unknown bitmap node %u\n", node_id);
}

#if 0

static void cb_b_layer_create(void *user_data, VNodeID node_id, VLayerID layer_id, const char *name, VNBLayerType type)
{
	NBitmap	*bit;

	if((bit = ndb_lookup(node_id)) != NULL)
	{
		printf(" Node %u has a layer called '%s' (or %u), type %u\n", node_id, name, layer_id, type);
		if(type != VN_B_LAYER_INTEGER8)
		{
			fprintf(stderr, "RENDER: Can't handle type %u bitmap layer\n", type);
			return;
		}
		if(strcmp(name, "color_r") == 0)
			bit->l_red   = layer_id;
		else if(strcmp(name, "color_g") == 0)
			bit->l_green = layer_id;
		else if(strcmp(name, "color_b") == 0)
			bit->l_blue  = layer_id;
		else
			return;
		vll_send_b_layer_subscribe(node_id, layer_id, BITMAP_LEVEL);
	}
}

static void cb_b_layer_set_tile(void *user_data, VNodeID node_id, VLayerID layer_id, uint8 level, uint16 tile_x, uint16 tile_y, void *data)
{
	NBitmap	*node;

	if((node = ndb_lookup(node_id)) != NULL)
	{
		uint32	x, y, ah, aw = 0, channel, tile = node->tile;
		uint8	*get, *put;

		if(layer_id == node->l_red)
			channel = 0;
		else if(layer_id == node->l_green)
			channel = 1;
		else if(layer_id == node->l_blue)
			channel = 2;
		else
		{
			fprintf(stderr, "RENDER: Got b_layer_set_tile for unknown layer %u\n", layer_id);
			return;
		}
		/* Copy <data> into client's copy of tile, for later passing on to OpenGL. */
		aw = MIN(1 << tile, node->width);
		ah = MIN(1 << tile, node->height);
		for(y = 0; y < ah; y++)
		{
			get = (uint8 *) data + y * (1 << tile);
			put = node->pixel + 3 * ((tile_y << tile) + y) * node->width + 3 * (tile_x << tile) + channel;
			for(x = 0; x < aw; x++, put += 3)
				*put = *get++;
		}
		if(!(node->dirtymap[tile_y * node->dirtymodulo + tile_x / CHAR_BIT] & (1 << (tile_x % CHAR_BIT))))
		{
			node->dirty++;
/*			printf("Tile (%u,%u): Set bit at row %u, byte %u, bit %u, perhaps?\n", tile_x, tile_y, tile_y * node->dirtymodulo, tile_x / CHAR_BIT, tile_x % CHAR_BIT);*/
			if(!g_list_find(global_data.dirty_bitmaps, node))
				global_data.dirty_bitmaps = g_list_append(global_data.dirty_bitmaps, node);
			node->dirtymap[tile_y * node->dirtymodulo + tile_x / CHAR_BIT] |= 1 << (tile_x % CHAR_BIT);
		}
/*
		glBindTexture(GL_TEXTURE_2D, node->gl_id);
#if !defined(USE_GL_TEXSUBIMAGE2D)
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, node->width, node->height, 0, GL_RGB, GL_UNSIGNED_BYTE, node->pixel);
#else
		for(y = 0; y < ah; y++)
		{
			get = node->pixel   + 3 * ((tile_y << tile) + y) * node->width + 3 * (tile_x << tile);
			put = node->tilebuf + 3 * y * (1 << tile);
			for(x = 0; x < aw; x++)
			{
				*put++ = *get++;
				*put++ = *get++;
				*put++ = *get++;
			}
		}
		printf("Setting subimage at (%u,%u), %ux%u pixels\n", tile_x << tile, tile_y << tile, 1 << tile, 1 << tile);
		glTexSubImage2D(GL_TEXTURE_2D, 0, tile_x << tile, tile_y << tile, 1 << tile, 1 << tile, GL_RGB, GL_UNSIGNED_BYTE, node->tilebuf);
		printf("Updated tile (%u,%u), layer %u\n", tile_x, tile_y, layer_id);
		global_data.dirty_tiles = TRUE;
#endif
*/
	}
}

#endif

/* Go through list of dirty bitmaps, and send them to OpenGL.
** FIXME: Since we know the number of tiles that are dirty in each bitmap, and also their positions (implicitly
**	  through the bitmap), it would probably make sense to do some kind of decision between always updating
**	  the entire bitmap, and only updating tiles using glTexSubImage2D().
*/
static void update_bitmaps(void)
{
	GList	*iter;
	NBitmap	*node;

	for(iter = global_data.dirty_bitmaps; iter != NULL; iter = g_list_next(iter))
	{
		node = iter->data;
		glBindTexture(GL_TEXTURE_2D, node->gl_id);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, node->width, node->height, 0, GL_RGB, GL_UNSIGNED_BYTE, node->pixel);
/*		printf("Node %lu has %u dirty tiles\n", node->head.id, node->dirty);*/
		node->dirty = 0U;
		memset(node->dirtymap, 0, node->dirtymodulo * (node->height / VN_B_TILE_SIZE));
	}
	g_list_free(global_data.dirty_bitmaps);
	global_data.dirty_bitmaps = NULL;
}

#if 0

/* ------------------------------------------------------------------------------------------------------------------------- */

static void cb_m_fragment_geometry_create(void *user, VNodeID node_id, VNMFragmentID id, const char *layer_r, const char *layer_g, const char *layer_b)
{
	NMaterial	*m;

	if((m = ndb_lookup(node_id)) != NULL)
	{
		GList	*iter;
		NObject	*obj;

		g_snprintf(m->map_u, sizeof m->map_u, "%s", layer_r);
		g_snprintf(m->map_v, sizeof m->map_v, "%s", layer_g);

		for(iter = m->users; iter != NULL; iter = g_list_next(iter))
		{
			obj = iter->data;
			obj->map_u = g_polygon_layer_find_by_name(obj->geometry, layer_r);
			obj->map_v = g_polygon_layer_find_by_name(obj->geometry, layer_g);
		}
	}
}

static void cb_m_fragment_texture_create(void *user, VNodeID node_id, VNMFragmentID id, VNodeID bitmap, const char * layer_r, const char * layer_g, const char * layer_b, VNMFragmentID mapping, VNMFragmentID warp)
{
	NMaterial	*m;

	if((m = ndb_lookup(node_id)) != NULL)
	{
		NBitmap	*bit;

		m->texture = bitmap;
		if((bit = ndb_lookup(bitmap)) == NULL)
		{
			ndb_b_new(bitmap);
			vll_send_subscribe(bitmap, V_NT_BITMAP);
		}
	}
/*	printf("Texture: id %u, with bitmap=%u, r='%s', g='%s', b='%s', mapping=%u, warp=%u\n", id, bitmap, layer_r, layer_g, layer_b, mapping, warp);*/
}

#endif

/* ------------------------------------------------------------------------------------------------------------------------- */

static void gcb_visibility(int state)
{
	global_data.visible = (state == GLUT_VISIBLE);
}

/* Move the avatar. If <spd_strafe> is non-zero, apply a strafing motion. If <spd_fly> is non-zero move
** in current direction. If <spd_vertical> is non-zero, move that much along the local vertical.
*/
static void move_avatar(real64 spd_strafe, real64 spd_vertical, real64 spd_fly)
{
	if(spd_strafe != 0.0)
	{
		real32	strafe_x, strafe_z;

		strafe_x =  cos(global_data.cam_yaw) * sin(global_data.cam_pitch);
		strafe_z = -sin(global_data.cam_yaw) * sin(global_data.cam_pitch);

		global_data.cam_pos[0] += spd_strafe * strafe_x;
		global_data.cam_pos[2] += spd_strafe * strafe_z;
	}
	if(spd_fly != 0.0)
	{
		global_data.cam_pos[0] += spd_fly * global_data.cam_fwd[0];
		global_data.cam_pos[1] += spd_fly * global_data.cam_fwd[1];
		global_data.cam_pos[2] += spd_fly * global_data.cam_fwd[2];
	}
	if(spd_vertical != 0.0)
	{
		real32	up[3];

		up[0] =  cos(global_data.cam_yaw + M_PI_2) * sin(global_data.cam_pitch - M_PI_2);
		up[1] =  cos(global_data.cam_pitch - M_PI_2);
		up[2] = -sin(global_data.cam_yaw + M_PI_2) * sin(global_data.cam_pitch - M_PI_2);
		global_data.cam_pos[0] += spd_vertical * up[0];
		global_data.cam_pos[1] += spd_vertical * up[1];
		global_data.cam_pos[2] += spd_vertical * up[2];
	}
	if(global_data.move_avatar)
		verse_send_o_transform_pos_real32(global_data.avatar, 0, 0, global_data.cam_pos, NULL, NULL, NULL, 0.0f);
}

static void rotate_avatar(double dyaw, double dpitch)
{
	global_data.cam_yaw   += dyaw;
	global_data.cam_pitch += dpitch;

	if(global_data.cam_pitch < DEG2RAD(1))
		global_data.cam_pitch = DEG2RAD(1);
	else if(global_data.cam_pitch > M_PI - DEG2RAD(1))
		global_data.cam_pitch = M_PI - DEG2RAD(1);

	if(global_data.move_avatar)
	{
		VNQuat32	rot;

		eulers_to_quat(&rot, global_data.cam_yaw - M_PI / 2.0, global_data.cam_pitch);
		verse_send_o_transform_rot_real32(global_data.avatar, 0, 0, &rot, NULL, NULL, NULL, 0.0f);
	}
}

#if 0
static void camera_to_object(void)
{
	NObject		*obj;
	NGeometry	*geo;
	real32		max[3];
	const NVLayer	*vtx;
	const VLDataXYZ	*xyz;
	unsigned int	i;

	if(global_data.current < 0)
		return;
	if((obj = (NObject *) ndb_lookup(global_data.current)) == NULL || obj->head.type != V_NT_OBJECT)
		return;
	if((geo = obj->geometry) == NULL)
		return;
	if((vtx = g_vertex_layer_find(geo, 0u)) == NULL)
		return;
	max[0] = max[1] = max[2] = -1E200f;
	xyz = vtx->data.xyz;
	for(i = 0; i < geo->num_vertices; i++)
	{
		if(xyz[i].x == V_REAL64_MAX ||
		   xyz[i].y == V_REAL64_MAX ||
		   xyz[i].z == V_REAL64_MAX)
			continue;
		if(fabs(xyz[i].x) > max[0])
			max[0] = fabs(xyz[i].x);
		if(fabs(xyz[i].y) > max[1])
			max[1] = fabs(xyz[i].y);
		if(fabs(xyz[i].z) > max[2])
			max[2] = fabs(xyz[i].z);
	}
	printf(" max: (%g,%g,%g)\n", max[0], max[1], max[2]);
	global_data.cam_pos[0] = obj->pos[0] + max[0] + 2.0;
	global_data.cam_pos[1] = obj->pos[1] + max[1] + 2.0;
	global_data.cam_pos[2] = obj->pos[2] + max[2] + 2.0;
	/* Fixme: need to compute yaw and pitch toward object center, too. */
}
#endif

static void move_object(real64 dx, real64 dy, real64 dz)
{
	NObject	*obj;

	if((obj = ndb_lookup(global_data.current)) != NULL)
	{
		real32	newpos[3];

		newpos[0] = obj->pos[0] + dx;
		newpos[1] = obj->pos[1] + dy;
		newpos[2] = obj->pos[2] + dz;
		verse_send_o_transform_pos_real32(obj->head.id, 0u, 0u, newpos, NULL, NULL, NULL, 0.0f);
	}
}

static void center_object(void)
{
	const NObject	*obj;

	if((obj = ndb_lookup(global_data.current)) != NULL)
	{
		const real32	origin[] = { 0.0f, 0.0f, 0.0f };

		verse_send_o_transform_pos_real32(obj->head.id, 0u, 0u, origin, NULL, NULL, NULL, 0.0f);
	}
}

static void send_ping(void)
{
	char	buf[32];
	uint32	s, f;

	verse_session_get_time(&s, &f);
	sprintf(buf, "%u.%u", s, f);
	verse_send_ping(global_data.server, buf);
}

static void move_vertices(real32 factor)
{
	NObject	*obj;
	NVLayer	*vl;
	uint32	i;

	if((obj = ndb_lookup(global_data.current)) == NULL)
	{
		printf("No current object -- select one in the RMB menu\n");
		return;
	}
	vl = obj->geometry->vertex_layers->data;
	for(i = 0; i < obj->geometry->num_vertices; i++)
		verse_send_g_vertex_set_xyz_real32(obj->geometry->head.id, vl->id, i,
						   factor * vl->data.xyz[i].x,
						   factor * vl->data.xyz[i].y,
						   factor * vl->data.xyz[i].z);
}

static void object_hidden_toggle(void)
{
	NObject	*obj;

	if((obj = ndb_lookup(global_data.current)) != NULL)
	{
		verse_send_o_hide(obj->head.id, !obj->hidden);
	}
}

static void buffer_history_toggle(BufferHistory *h)
{
	h->enabled = !h->enabled;
	if(h->enabled)
	{
		memset(&h->history, 0, sizeof h->history);
		h->sum = h->count = 0;
	}
}

static void buffer_history_update(BufferHistory *h)
{
	size_t	now = verse_session_get_size();

	if(!h->enabled)
		return;

	h->sum += now;
	if(++h->count == 10)
	{
		size_t	i;

		for(i = 0; i < sizeof h->history / sizeof *h->history - 1; i++)
			h->history[i] = h->history[i + 1];
		h->history[i] = h->sum / h->count;
		h->sum = h->count = 0;
	}
}

static void buffer_history_display(const BufferHistory *h)
{
	size_t	i, max = 0;
	GLfloat	x;

	if(!h->enabled)
		return;

	for(i = 0; i < sizeof h->history / sizeof *h->history; i++)
		if(h->history[i] > max)
			max = h->history[i];
	if(max == 0)
		max = 1;
	glBegin(GL_LINES);
	glColor3f(0.0f, 0.8f, 0.3f);
	for(i = 0, x = 30.0; i < sizeof h->history / sizeof *h->history; i++, x += 1.0f)
	{
		glVertex2f(x, 15.0f);
		glVertex2f(x, 16.0f + 25.0f * h->history[i] / max);
	}
	glEnd();
}

static void polygon_count_display(void)
{
	char	buf[8], *ptr;
	int	len;

	len = snprintf(buf, sizeof buf, "%u", global_data.polygon_count);
	glRasterPos2f(global_data.win_width - 30.0 - 8.0 * len, 15.0f);
	for(ptr = buf; *ptr;)
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *ptr++);
}

static void gcb_keyboard(unsigned char key, int x, int y)
{
	if(global_data.loud_keys)
		printf("Ordinary key: %d\n", (int) key);

	switch(key)
	{
		case 'w':
			move_avatar( 0.0, 0.0, 0.3);
			break;
		case 'a':
			move_avatar(-0.3, 0.0, 0.0);
			break;
		case 's':
			move_avatar( 0.0, 0.0, -0.3);
			break;
		case 'd':
			move_avatar( 0.3, 0.0, 0.0);
			break;
		case 'W':
			move_avatar(0.0, 0.3, 0.0);
			break;
		case 'S':
			move_avatar(0.0, -0.3, 0.0);
			break;

		case 'u':
			move_object(0, 0, -MOVE_DELTA);
			break;
		case 'j':
			move_object(0, 0, MOVE_DELTA);
			break;
		case 'h':
			move_object(-MOVE_DELTA, 0, 0);
			break;
		case 'k':
			move_object(MOVE_DELTA, 0, 0);
			break;

		case 'U':
			move_object(0, 0, -1E4 * MOVE_DELTA);
			break;
		case 'J':
			move_object(0, 0, 1E4 * MOVE_DELTA);
			break;
		case 'H':
			move_object(-1E4 * MOVE_DELTA, 0, 0);
			break;
		case 'K':
			move_object(1E4 * MOVE_DELTA, 0, 0);
			break;

		case '0':
			center_object();
			break;

		case 'p':
			send_ping();
			break;
		case 't':
			global_data.texture_active = !global_data.texture_active;
			break;
		case 'T':
			global_data.wireframe_active = !global_data.wireframe_active;
			break;
		case 'g':
			global_data.ground_active = !global_data.ground_active;
			break;
		case 'G':
			global_data.axis_active = !global_data.axis_active;
			break;
		case 'l':
			global_data.lights_active ^= 1;
			break;
		case 'z':
			global_data.object_cubes_active ^= 1;
			break;
		case 'Z':
			global_data.culling_active ^= 1;
			break;

		case 'x':
			global_data.move_avatar = !global_data.move_avatar;
			break;

		case 'c':
			global_data.show_time ^= 1;
			break;
		case 'C':
			global_data.show_camera ^= 1;
			break;
		case 'v':
			global_data.show_vertices ^= 1;
			break;
		case 'V':
			global_data.show_uvs ^= 1;
			break;
		case 'P':
			global_data.show_polygons ^= 1;
			break;
		
		case '+':
			move_vertices(1.2);
			break;
		case '-':
			move_vertices(0.8);
			break;

		case ':':
			object_hidden_toggle();
			break;

/*		case ',':
			camera_to_object();
			break;
*/		case '.':
			buffer_history_toggle(&global_data.buf_hist);
			break;

		case 13:
			if(global_data.mouse_modify)
				global_data.mouse_button ^= 1;
			break;

		case 'Q':
		case 27:
			verse_send_connect_terminate(global_data.server, "User quit");
			do
				verse_callback_update(10000);
			while(verse_session_get_size() > 10);
			exit(0);
			break;
		case 127:		/* Hopefully, this is everybody's delete key. */
			if(global_data.current != ~0u)
			{
				NObject	*obj = ndb_lookup(global_data.current);
				if(obj != NULL)
				{
					if(obj->geometry != NULL)
						verse_send_node_destroy(obj->geometry->head.id);
					if(obj->material != NULL)
						verse_send_node_destroy(obj->material->head.id);
					verse_send_node_destroy(global_data.current);
				}
			}
			break;
	}
}

static void gcb_special(int key, int x, int y)
{
	if(global_data.loud_keys)
		printf("Special key:  %d\n", key);

	switch(key)
	{
		case GLUT_KEY_UP:
			move_avatar( 0.0, 0.0, 0.3);
			break;
		case GLUT_KEY_LEFT:
			move_avatar(-0.3, 0.0, 0.0);
			break;
		case GLUT_KEY_DOWN:
			move_avatar( 0.0, 0.0, -0.3);
			break;
		case GLUT_KEY_RIGHT:
			move_avatar( 0.3, 0.0, 0.0);
			break;
	}
}

static void gcb_menu(int value)
{
	if((value & (MENU_MASK)) == MENU_CAMERA)
		global_data.cam_mode = value & ((1 << 16) - 1);
	else
		global_data.current = value;
}

static void lights_begin(void)
{
	GList	*iter;
	GLint	i, n;

	if(!global_data.lights_active)
		return;

	glGetIntegerv(GL_MAX_LIGHTS, &n);
	if(n == 0)
		return;

	for(i = 0, iter = global_data.objects; i < n && iter != NULL; iter = g_list_next(iter))
	{
		NObject	*obj = iter->data;

		if(obj->light[0] == 0.0 && obj->light[1] == 0.0 && obj->light[2] == 0.0)
			continue;
		glLightfv(GL_LIGHT0 + i, GL_POSITION, obj->pos);
		glLightfv(GL_LIGHT0 + i, GL_DIFFUSE, obj->light);
		glLightfv(GL_LIGHT0 + i, GL_SPECULAR, obj->light);
		glEnable(GL_LIGHT0 + i);
		i++;
	}
	glEnable(GL_LIGHTING);
	glEnable(GL_AUTO_NORMAL);
}

static void lights_end(void)
{
	glDisable(GL_LIGHTING);
	glDisable(GL_AUTO_NORMAL);
}

static void gcb_display(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if(global_data.dirty_bitmaps)
		update_bitmaps();

	buffer_history_update(&global_data.buf_hist);

	/* Recompute camera's forward vector. */
	global_data.cam_fwd[0] =  cos(global_data.cam_yaw + M_PI_2) * sin(global_data.cam_pitch);
	global_data.cam_fwd[1] =  cos(global_data.cam_pitch);
	global_data.cam_fwd[2] = -sin(global_data.cam_yaw + M_PI_2) * sin(global_data.cam_pitch);

	glPushMatrix();
	switch(global_data.cam_mode)
	{
		case MENU_CAMERA_FREEFLIGHT:
			gluLookAt(global_data.cam_pos[0], global_data.cam_pos[1], global_data.cam_pos[2],
				  global_data.cam_pos[0] + global_data.cam_fwd[0], global_data.cam_pos[1] + global_data.cam_fwd[1],
				  global_data.cam_pos[2] + global_data.cam_fwd[2],	0.0, 1.0, 0.0);
			break;
		case MENU_CAMERA_ISOMETRIC:
			gluLookAt(15.0, 8.0, 1500.0,  0.0, 0.0, 0.0,  0.0, 1.0, 0.0);
			break;
		case MENU_CAMERA_SATELLITE:
			gluLookAt(0.0, 2000.0, 0.0,  0.0, 0.0, 0.0,  0.0, 0.0, -1.0);
			break;
		case MENU_CAMERA_OBJECT:
			{
				NObject	*obj;

				if((obj = ndb_lookup(global_data.current)) != NULL)
				{
					real32	x = obj->pos[0], y = obj->pos[1], z = obj->pos[2], fwd[3];

					ndb_o_get_forward(obj, fwd);
					gluLookAt(x, y, z, 
						  x + fwd[0], y + fwd[1], z + fwd[2],
						  0.0, 1.0, 0.0);
				}
			}
			break;
	}

	/* A ground plane, for reference. */
	if(global_data.ground_active)
	{
		glColor3ub(0x30, 0x20, 0x60);
		glBegin(GL_QUADS);
		 glVertex3f(-10.0f, -0.5f, -10.0f);
		 glVertex3f( 10.0f, -0.5f, -10.0f);
		 glVertex3f( 10.0f, -0.5f,  10.0f);
		 glVertex3f(-10.0f, -0.5f,  10.0f);
		glEnd();
	}

	/* Let's have a nice set of coordinate axis. Very handy sometimes. */
	if(global_data.axis_active)
	{
		glBegin(GL_LINES);
		 glColor3f(1.0f, 0.0f, 0.0f);
		 glVertex3f(0.0f, 0.0f, 0.0f);
		 glVertex3f(10.0f, 0.0f, 0.0f);

		 glColor3f(0.0f, 1.0f, 0.0f);
		 glVertex3f(0.0f, 0.0f, 0.0f);
		 glVertex3f(0.0f, 10.0f, 0.0f);

		 glColor3f(0.0f, 0.0f, 1.0f);
		 glVertex3f(0.0f, 0.0f, 0.0f);
		 glVertex3f(0.0f, 0.0f, 10.0f);
		glEnd();
		glRasterPos3f(0.0, 0.0, 10.0f);
		glutBitmapCharacter(GLUT_BITMAP_9_BY_15, 'Z');
		glColor3f(0.0f, 1.0f, 0.0f);
		glRasterPos3f(0.0, 10.0, 0.0f);
		glutBitmapCharacter(GLUT_BITMAP_9_BY_15, 'Y');
		glColor3f(1.0f, 0.0f, 0.0f);
		glRasterPos3f(10.0f, 0.0f, 0.0f);
		glutBitmapCharacter(GLUT_BITMAP_9_BY_15, 'X');
	}
	lights_begin();
	ndb_render();
	lights_end();
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0.0f, global_data.win_width, 0.0f, global_data.win_height, -1.0f, 1.0f);
	buffer_history_display(&global_data.buf_hist);
	polygon_count_display();
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	global_data.frames++;
	glutSwapBuffers();
}

static void gcb_reshape(int width, int height)
{
	global_data.win_width  = width;
	global_data.win_height = height;
	global_data.cam_active = FALSE;

	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluPerspective(45, (GLfloat) width / height, 0.01f, 1000.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

static void gcb_timer(int value)
{
	verse_callback_update(1000);

	if(global_data.move_active && global_data.cam_mode == MENU_CAMERA_FREEFLIGHT)
		move_avatar(0.0, 0.0, global_data.move_speed);
	glutPostRedisplay();
	glutTimerFunc(global_data.timer_period, gcb_timer, 0);
}

static void gcb_motion(int x, int y)
{
	if(global_data.cam_active)
	{
		gint	dx, dy;

		dx = x - global_data.last_x;
		dy = y - global_data.last_y;

		rotate_avatar(-DEG2RAD(dx) / 4.0, DEG2RAD(dy) / 4.0);

		global_data.last_x = x;
		global_data.last_y = y;
	}
	else if(global_data.move_active)
		global_data.move_speed = (y - global_data.last_y) / -100.0;
}

static void gcb_mouse(int button, int state, int x, int y)
{
	if(global_data.mouse_modify)
		button = global_data.mouse_button;

	if(button == GLUT_LEFT_BUTTON)
	{
		if(global_data.cam_mode == MENU_CAMERA_FREEFLIGHT)
			global_data.cam_active = FALSE;
		global_data.move_active = (state == GLUT_DOWN);
	}
	else if(button == GLUT_MIDDLE_BUTTON)
		global_data.cam_active = (state == GLUT_DOWN);
	else if(button == 3)
		move_avatar(0.0, 0.0, 1.0);
	else if(button == 4)
		move_avatar(0.0, 0.0, -1.0);
	global_data.last_x = x;
	global_data.last_y = y;
}

static void build_camera_menu(void)
{
	global_data.cam_menu = glutCreateMenu(gcb_menu);
	glutAddMenuEntry("Free Flying",	MENU_CAMERA | MENU_CAMERA_FREEFLIGHT);
	glutAddMenuEntry("Isometric",	MENU_CAMERA | MENU_CAMERA_ISOMETRIC);
	glutAddMenuEntry("Satellite",	MENU_CAMERA | MENU_CAMERA_SATELLITE);
	glutAddMenuEntry("Object",	MENU_CAMERA | MENU_CAMERA_OBJECT);
}

static void init_glut(int *argcp, char **argv)
{
	glutInit(argcp, argv);

	glutInitWindowSize(global_data.win_width, global_data.win_height);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);

	global_data.window = glutCreateWindow("Verse Simple Renderer");
	global_data.menu = glutCreateMenu(gcb_menu);
	global_data.obj_menu = glutCreateMenu(gcb_menu);
	build_camera_menu();
	glutSetMenu(global_data.menu);
	glutAddSubMenu("Objects", global_data.obj_menu);
	glutAddSubMenu("Camera", global_data.cam_menu);
	glutAttachMenu(GLUT_RIGHT_BUTTON);

	glutKeyboardFunc(gcb_keyboard);
	glutSpecialFunc(gcb_special);
	glutDisplayFunc(gcb_display);
	glutVisibilityFunc(gcb_visibility);
	glutTimerFunc(global_data.timer_period, gcb_timer, 0);
	glutReshapeFunc(gcb_reshape);
	glutMotionFunc(gcb_motion);
	glutMouseFunc(gcb_mouse);

	glFrontFace(GL_CW);
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
}

static void cb_exit(void)
{
	int	t2;
	real32	s;

	t2 = glutGet(GLUT_ELAPSED_TIME);
	s  = 1E-3 * (t2 - global_data.start_time);
	printf("Drew %u frames in %.2f s => %.2f fps\n", global_data.frames, s, global_data.frames / s);

	if(global_data.session)
	{
		verse_session_destroy(global_data.session);
		global_data.session = NULL;
	}
}

/* ------------------------------------------------------------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
	boolean		gl_info = FALSE;
	int		i;

	global_data.server = "localhost";
	global_data.home_avatar = FALSE;
	global_data.home_pos[0] = 7.0f;
	global_data.home_pos[1] = 16.0f;
	global_data.home_pos[2] = 13.0f;
	global_data.home_yaw    = DEG2RAD(20);
	global_data.home_pitch  = DEG2RAD(125);
	global_data.move_avatar = TRUE;
	global_data.loud_keys = FALSE;
	global_data.mouse_modify = FALSE;
	global_data.mouse_button = 0;
	global_data.timer_period = 35;
	global_data.win_width  = 640;
	global_data.win_height = 480;
	global_data.visible = TRUE;
	global_data.show_vertices = FALSE;
	global_data.show_polygons = FALSE;
	global_data.show_uvs = FALSE;
	global_data.lights_active = FALSE;

	for(i = 1; argv[i] != NULL; i++)
	{
		if(strncmp(argv[i], "-ip=", 4) == 0)
			global_data.server = argv[i] + 4;
		else if(strcmp(argv[i], "-tiny") == 0)
		{
			global_data.win_width  = 240;
			global_data.win_height = 320;
		}
		else if(strcmp(argv[i], "-glinfo") == 0)
			gl_info = TRUE;
		else if(strncmp(argv[i], "-home=", 6) == 0)
		{
			if(sscanf(argv[i] + 6, "%f,%f,%f", &global_data.home_pos[0], &global_data.home_pos[1], &global_data.home_pos[2]) == 3)
				global_data.home_avatar = TRUE;
			else
				fprintf(stderr, "render: Couldn't read three floats for -home, expected x,y,z\n");
		}
		else if(strncmp(argv[i], "-look=", 6) == 0)
		{
			if(sscanf(argv[i] + 6, "%f,%f", &global_data.home_yaw, &global_data.home_pitch) == 2)
			{
				global_data.home_yaw = DEG2RAD(global_data.home_yaw);
				global_data.home_pitch = DEG2RAD(global_data.home_pitch);
				global_data.home_avatar = TRUE;
			}
			else
				fprintf(stderr, "render: Couldn't read two floats for -look, expected yaw,pitch\n");
		}
		else if(strcmp(argv[i], "-loudkeys") == 0)
			global_data.loud_keys = TRUE;
		else if(strcmp(argv[i], "-modmouse") == 0)
			global_data.mouse_modify = TRUE;
		else if(strstr(argv[i], "help") != NULL)
		{
			printf("Usage: %s [-ip=HOST[:PORT]] [-tiny] [glinfo] [-home=x,y,z] [-loudkeys] [-modmouse]\n", argv[0]);
			printf("Default HOST and PORT are localhost and 4950, respectively.\n");
			return EXIT_SUCCESS;
		}
		else
			fprintf(stderr, "**Unknown option \"%s\" ignored\n", argv[i]);
	}

	init_glut(&argc, argv);

	if(gl_info)
	{
		printf("OpenGL    version: %s\n", glGetString(GL_VERSION));
		printf("OpenGL     vendor: %s\n", glGetString(GL_VENDOR));
		printf("OpenGL extensions: %s\n", glGetString(GL_EXTENSIONS));
	}

	ndb_init();

	if((global_data.session = verse_send_connect("render", "<nothing>", global_data.server, NULL)) != NULL)
	{
		atexit(cb_exit);

		verse_callback_set((void *) verse_send_connect_accept,		(void *) cb_connect_accept,	NULL);
		verse_callback_set((void *) verse_send_ping,			(void *) cb_ping,		NULL);
		verse_callback_set((void *) verse_send_node_create,		(void *) cb_node_create,	NULL);
		verse_callback_set((void *) verse_send_node_name_set,		(void *) cb_node_name_set,	NULL);
		verse_callback_set((void *) verse_send_node_destroy,		(void *) cb_node_destroy,	NULL);
		verse_callback_set((void *) verse_send_o_link_set,		(void *) cb_o_link_set,		NULL);
		verse_callback_set((void *) verse_send_o_transform_pos_real32,	(void *) cb_o_transform_pos_real32,	NULL);
		verse_callback_set((void *) verse_send_o_transform_rot_real32,	(void *) cb_o_transform_rot_real32,	NULL);
		verse_callback_set((void *) verse_send_o_transform_scale_real32, (void *) cb_o_transform_scale_real32,	NULL);
		verse_callback_set((void *) verse_send_o_light_set,		(void *) cb_o_light_set,		NULL);
		verse_callback_set((void *) verse_send_o_hide,			(void *) cb_o_hide, NULL);

		verse_callback_set((void *) verse_send_g_layer_create,		(void *) cb_g_layer_create,		NULL);
		verse_callback_set((void *) verse_send_g_vertex_set_xyz_real64,	(void *) cb_g_vertex_set_xyz_real64,	NULL);
		verse_callback_set((void *) verse_send_g_vertex_set_xyz_real32,	(void *) cb_g_vertex_set_xyz_real32,	NULL);
		verse_callback_set((void *) verse_send_g_vertex_delete_real32,	(void *) cb_g_vertex_delete,		NULL);
		verse_callback_set((void *) verse_send_g_vertex_delete_real64,	(void *) cb_g_vertex_delete,		NULL);
		verse_callback_set((void *) verse_send_g_polygon_set_corner_uint32, (void *) cb_g_polygon_set_corner_uint32, NULL);
		verse_callback_set((void *) verse_send_g_polygon_set_corner_real64, (void *) cb_g_polygon_set_corner_real64,	NULL);
		verse_callback_set((void *) verse_send_g_polygon_delete,	(void *) cb_g_polygon_delete,		NULL);
		verse_callback_set((void *) verse_send_g_bone_create,		(void *) cb_g_bone_create,		NULL);

		verse_callback_set((void *) verse_send_b_dimensions_set,	(void *) cb_b_dimensions_set,		NULL);
/*		verse_callback_set((void *) verse_send_b_layer_create,		(void *) cb_b_layer_create,		NULL);
		verse_callback_set((void *) verse_send_b_layer_set_tile,	(void *) cb_b_layer_set_tile,		NULL);

		verse_callback_set((void *) verse_send_m_fragment_geometry_create, (void *) cb_m_fragment_geometry_create, NULL);
		verse_callback_set((void *) verse_send_m_fragment_texture_create, (void *) cb_m_fragment_texture_create,	NULL);
*/
		global_data.start_time = glutGet(GLUT_ELAPSED_TIME);
		global_data.frames = 0;
		global_data.visible = TRUE;

		global_data.show_time = FALSE;

		global_data.yaw = 0.0;
		global_data.current = ~0u;
		global_data.last_x  = global_data.last_y = -1;

		global_data.cam_mode    = MENU_CAMERA_FREEFLIGHT;
		global_data.cam_pos[0]  = global_data.home_pos[0];
		global_data.cam_pos[1]  = global_data.home_pos[1];
		global_data.cam_pos[2]  = global_data.home_pos[2];
		global_data.cam_pitch   = global_data.home_pitch;
		global_data.cam_yaw     = global_data.home_yaw;
		global_data.move_active = FALSE;

		global_data.ping_active = FALSE;

		global_data.object_cubes_active = TRUE;
		global_data.wireframe_active = FALSE;
		global_data.texture_active   = FALSE;
		global_data.axis_active	     = TRUE;
		global_data.ground_active    = TRUE;
		global_data.lights_active    = FALSE;
		global_data.culling_active   = TRUE;

		global_data.buf_hist.enabled = TRUE;
		global_data.buf_hist.sum = global_data.buf_hist.count = 0;
		memset(global_data.buf_hist.history, 0, sizeof global_data.buf_hist.history);

		glutMainLoop();
	}
	return 0;
}
