#! /bin/env python
#
# A small test program to test tags, and setting largish numbers of them.
# Written in April 2005 by Emil Brink. Brief usage instructions:
#
# 1) Start a Verse server
# 2) Run this script, with arguments "-create -n=15000", changing the
#    number to however many tags you want to create
# 3) Leave the script from 2) running, since tags are created on its avatar
# 4) In a separate shell, start the script again, with no arguments
# 5) Once it seems to not be receiving any further tags, terminate it (by
#    pressing Control-C on most platforms)
# 6) The client prints out the number of tag_creates it received, this
#    number should match whatever you gave as argument to the first instance.
#

import sys
import verse as v

create    = False
amount    = 0
my_avatar = ~0

tag_groups = { }

def key(node_id, group_id):
	return str(node_id) + "." + str(group_id)

def cb_tag_create(node_id, group_id, tag_id, name, type, value):
	global tag_groups

	print "Tag", str(group_id) + "." + str(tag_id) + ",", name, type, value, "created"
	try:
		tg = tag_groups[key(node_id, group_id)]
	except:
		print "Unknown group, not storing"
		return
	tg[tag_id] = (name, type, value)

def cb_tag_group_create(node_id, group_id, name):
	global my_avatar, tag_groups
	print "Node", node_id, "got tag group", group_id, ":", name, "I am", my_avatar
	v.send_tag_group_subscribe(node_id, group_id)
	tag_groups[key(node_id, group_id)] = { }
	if create and node_id == my_avatar:
		print "Create-mode, creating", amount, "tags"
		for i in range(0, amount):
			v.send_tag_create(node_id, group_id, ~0, "test-tag-" + str(i), v.TAG_REAL64, 1.55 / float(i + 1))

def cb_node_create(node_id, type, owner):
	global my_avatar
	print "There's a node called", node_id
	if type == v.OBJECT and node_id != my_avatar:
		v.send_node_subscribe(node_id)

def cb_connect_accept(avatar, address, host_id):
	global my_avatar
	print
	print "Connected alright. Listing objects (for simplicity)."
	my_avatar = avatar
	v.send_node_list(1 << v.OBJECT)
	v.send_node_name_set(my_avatar, "tagtest")
	v.send_node_subscribe(my_avatar)
	if create:
		print "Creating test tag group on my avatar,", my_avatar
		v.send_tag_group_create(my_avatar, ~0, "test")

if __name__ == "__main__":
	v.callback_set(v.SEND_CONNECT_ACCEPT,	cb_connect_accept)
	v.callback_set(v.SEND_NODE_CREATE,	cb_node_create)
	v.callback_set(v.SEND_TAG_GROUP_CREATE,	cb_tag_group_create)
	v.callback_set(v.SEND_TAG_CREATE,	cb_tag_create)

	ip = "localhost"
	for a in sys.argv:
		if a.startswith("-ip="):
			ip = a[4:]
		elif a.startswith("-n="):
			amount = int(a[3:])
		elif a == "-create":
			create = True

	if create:
		print "Create-mode enabled, amount=", amount
	v.send_connect("tagtest", "<secret>", ip, 0)

	try:
		while True:
			v.callback_update(10000)
	except KeyboardInterrupt:
		if not create:
			print "Aborting, here's what we got:"
			for g in tag_groups:
				print g, ":", len(tag_groups[g]), "tags"
		print "Disconnecting"
		v.send_connect_terminate(ip, "All done, bye bye")
		while v.session_get_size() > 10:
			v.callback_update(10000)
