#!/usr/bin/env python
#
# A simple little script to help start a Verse server, and set the
# options properly.
#
# Copyright (C) 2006 by PDC, KTH. Written by Emil Brink. Released under
# the FreeBSD license. See <http://www.freebsd.org/copyright/freebsd-license.html>
# for the license text, but be aware that this software has nothing to do
# with the FreeBSD project.
#
# More information about Verse can be found at <http://verse.blender.org/>.
#

VERSION = "0.2"

import os
import sys
import string

def import_or_die(module, as = None, req = None):
	"""A simple function to do a 'safe' import, with error reporting."""
	if as == None:
		s = "global %s" % module
	else:
		s = "global %s" % as
	s += " ; import %s" % module
	if as != None:	s += " as %s" % as
	if req != None:	s += " ; %s.require(\"%s\")" % (module, req)
	try:
		exec(s)
	except:
		print "Couldn't import Python module '%s', which is required for this to" % module
		print "run properly. Please see documentation for more information. Press"
		print "return to exit."
		raw_input()
		sys.exit(1)

def get_server_version(server = "verse"):
	try:
		pipe = os.popen(server + " -version", "r")
		ver = pipe.read()
		return ver.strip()
	except:
		return "(unknown)"
	

import_or_die("pygtk", req = "2.0")
import_or_die("gobject")
import_or_die("gtk")

class VerseStarter:

	def _start_server(self):
		"""Builds command line, and launches the Verse server process."""
		args = [self.verse]
		args += ["-port=%u" % int(self.port.get_value())]
		if self.msreg.get_active():
			args += ["-ms"]
			ip = self.ms_ip.get_text()
			if ip != "master.uni-verse.org:4950":
				args += ["-ms:ip=%s" % ip]
			desc = self.ms_desc.get_text()
			if len(desc) > 0:
				args += ["-ms:de=\"%s\"" % desc]
			tags = self.ms_tags.get_text()
			if len(tags) > 0:
				args += ["-ms:ta=%s" % tags]
		s = args[0]
		for a in args[1:]: s += " %s" % a
		print "Starting server: %s" % s
		try:
			self.pid = os.execv(self.verse, args)
		except Exception:
			print "Unable to start server"

	def _row(self, label, right):
		hbox = gtk.HBox(False)
		hbox.pack_start(gtk.Label(label), True)
		self.sizegroup.add_widget(right)
		hbox.pack_start(right, False)
		return hbox

	def _evt_find_clicked(self, wid, entry):
		old = self.verse

		def _evt_response(wid, response):
			if response == gtk.RESPONSE_ACCEPT:
				self.verse = wid.get_filename()

		if self.chooser == None:
			self.chooser = gtk.FileChooserDialog(title = "Find Verse Server Binary")
			self.chooser.add_buttons(gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
						gtk.STOCK_OK, gtk.RESPONSE_ACCEPT)
			self.chooser.connect("response", _evt_response)
		self.chooser.show()
		self.chooser.grab_add()
		self.chooser.run()
		self.chooser.grab_remove()
		self.chooser.hide()
		if self.verse != old:
			entry.set_text(get_server_version(self.verse))

	def _evt_port_default_clicked(self, wid):
		self.port.set_value(4950)

	def _evt_ms_toggled(self, wid):
		on = wid.get_active()
		self.ms_ip.set_sensitive(on)
		self.ms_desc.set_sensitive(on)
		self.ms_tags.set_sensitive(on)
		if on:
			self.ms_desc.grab_focus()

	def _evt_quit_clicked(self, wid):
		self.do_run = False
		gtk.main_quit()

	def _evt_start_clicked(self, wid):
		self.do_run = True
		gtk.main_quit()

	def _evt_win_delete(self, wid, evt):
		self.do_run = False
		gtk.main_quit()
		

	def __init__(self, verse):
		self.verse = verse
		self.pid = None
		self.do_run = False
		self.chooser = None

		self.win = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.win.connect("delete_event", self._evt_win_delete)
		self.win.set_title("Verse Server Starter")
		self.sizegroup = gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)
		vbox = gtk.VBox(False, 0)

		hbox = gtk.HBox(False, 0)
		ve = gtk.Entry()
		ve.set_text(get_server_version(self.verse))
		ve.set_editable(False)
		hbox.pack_start(ve, True)
		vb = gtk.Button(stock = gtk.STOCK_FIND)
		vb.connect("clicked", self._evt_find_clicked, ve)
		hbox.pack_start(vb)
		vbox.pack_start(self._row("Verse Version:", hbox), False)

		hbox = gtk.HBox(False, 0)
		self.port = gtk.SpinButton(gtk.Adjustment(4950, 0, 65535, 1, 1))
		hbox.pack_start(self.port, True, True, 0)
		db = gtk.Button("Default")
		db.connect("clicked", self._evt_port_default_clicked)
		hbox.pack_start(db, False, False, 0)
		vbox.pack_start(self._row("Port:", hbox), False)

		msfr = gtk.Frame("Master Server")
		vb = gtk.VBox(False, 0)
		hbox = gtk.HBox(False, 0)
		self.msreg = gtk.CheckButton("Register with Master Server?")
		self.msreg.connect("toggled", self._evt_ms_toggled)
		hbox.pack_start(self.msreg)
		vb.pack_start(hbox, False)
		self.ms_ip = gtk.Entry()
		self.ms_ip.set_text("master.uni-verse.org:4950")
		vb.pack_start(self._row("Master Server:", self.ms_ip), False)
		self.ms_desc = gtk.Entry()
		vb.pack_start(self._row("Description:", self.ms_desc), False)
		self.ms_tags = gtk.Entry()
		vb.pack_start(self._row("Tags:", self.ms_tags), False)
		self.msreg.toggled()
		msfr.add(vb)
		vbox.pack_start(msfr, True, True, 0)
		hbox = gtk.HBox(True, 0)
		qb = gtk.Button("Quit")
		qb.connect("clicked", self._evt_quit_clicked)
		hbox.pack_start(qb)
		sb = gtk.Button("Start Server")
		sb.connect("clicked", self._evt_start_clicked)
		sb.set_flags(gtk.CAN_DEFAULT)
		hbox.pack_start(sb)
		vbox.pack_start(hbox, False)
		self.win.add(vbox)
		sb.grab_default()

	def run(self):
		self.win.show_all()
		gtk.main()
		if self.do_run:
			self._start_server()

def main(args):
	if args == None:
		args = sys.argv[1:]
	verse = "../verse/verse"	# This is only sensible on some Linux systems, but at least then.
	for a in args:
		if a == "-h":
			print "Usage: verse-starter.py [-verse=BINARY] [-h] [-v]"
			print "Where BINARY is the fully-qualified name of a Verse server binary."
			print " -h  Prints this help text."
			print " -v  Prints version of this program, and exits."
			sys.exit(0)
		elif a == "-v":
			print VERSION
			sys.exit(0)
		elif a.startswith("-verse="):
			verse = a[7:]
	app = VerseStarter(verse)
	app.run()


if __name__ == "__main__":
	main(sys.argv)
